from .helper_functions import get_at_node_group, get_node, get_input, get_mod_group, get_mat_control_node,\
    get_biome_vars
from ..weight_layers_addon.WL_functions import get_node_by_type
from ..constants import ADDON_VERSION, NODE_NAMES, SAVED_BIOME_VERSION
import bpy


# UI
def draw_input_group(context, col, node, group: str = "", alert: list = []):
    """Draws a group of node inputs. The inputs are filtered according to the 'group' parameter.
    It looks for inputs that have their value hidden, and if the name matches 'group', it draws
    all inputs untill the next hidden input (unless that input has no name, when it will just put a spacer)"""
    group += "" if group.endswith("_") else "_"
    for socket in node.inputs:
        if socket.name.startswith(group) and all([not socket.links, socket.enabled, hasattr(socket, 'draw')]):
            row = col.row(align=True)

            if socket.hide_value:
                col.separator()
                continue

            if socket in alert:
                row.alert = socket.default_value == 0
            socket.draw(context, row, node, socket.name.replace(group, "").capitalize())
    # draw = False if group else True
    # for socket in node.inputs:
    #     if socket.hide_value:
    #         if socket.name == group:
    #             draw = True
    #         elif socket.name:
    #             draw = False
    #         else:
    #             col.separator()
    #         continue
    #     if all([not socket.links, socket.enabled, hasattr(socket, 'draw')]) and draw:
    #         row = col.row(align=True)

    #         if socket in alert:
    #             row.alert = socket.default_value == 0
    #         socket.draw(context, row, node, socket.name.capitalize())
    return


# Nodes
def organise_list_group(group):
    """ This organises the top level node group for each object
    (a.k.a the node tree used by the geo nodes modifier), so that the tree
    looks good."""

    nodes = group.nodes
    input_node = get_node_by_type(group, "GROUP_INPUT")
    output_node = get_node_by_type(group, "GROUP_OUTPUT")
    join_node = nodes.get(NODE_NAMES["join"])
    slot_nodes = []

    for n in nodes:
        if n.type == "DISTRIBUTE_POINTS_ON_FACES" or n.type == "GROUP":
            slot_nodes.append(n)
            n.location = (-160, -int(n.name) * 60)
            n.hide = True
            n.width = 220
            n.select = True if n == nodes.active else False
            n.use_custom_color = True
            n.color = (0.1, 0.7, 1)

    y_level = (len(slot_nodes) - 1) * 30 * -1 + 40
    output_node.location = (260, y_level)
    input_node.location = (-340, y_level)
    join_node.location = (100, y_level)


def setup_tree_group(mod_group: bpy.types.NodeTree):
    """This sets up the node group for individual alpha trees
    (a.k.a the controls for scale, width and z-height)"""
    tree_group = get_at_node_group(NODE_NAMES["tree"], "TREE")

    nodes = mod_group.nodes
    links = mod_group.links

    input_node = get_node_by_type(mod_group, "GROUP_INPUT")
    output_node = get_node_by_type(mod_group, "GROUP_OUTPUT")
    tree_node = nodes.new("GeometryNodeGroup")

    tree_node.node_tree = tree_group
    tree_node.name = tree_group.name
    tree_node.use_custom_color = True
    tree_node.color = (0.32, 0.78, 0.41)

    output_node.location = (-20, 0)
    tree_node.location = (-180, 0)

    in_socket = input_node.outputs.get("Geometry")
    out_socket = output_node.inputs.get("Geometry")

    link = links.new(in_socket, tree_node.inputs["Geometry"])
    links.new(tree_node.outputs["Geometry"], out_socket)


def swap_nodes(node, node_group=None):
    """This toggles the given node between the biome settings node group
    and the default point distribute node"""
    mod_group = node.id_data
    nodes = mod_group.nodes
    links = mod_group.links

    name = node.name
    type = node.type
    if node:
        input_socket = node.inputs[0].links[0].from_socket
        output_socket = node.outputs[0].links[0].to_socket
    else:
        input_node = get_node_by_type(mod_group, "GROUP_INPUT")
        join_node = nodes.get(NODE_NAMES["join"])
        input_socket = input_node.outputs[0]
        output_socket = join_node.inputs[0]

    if type == "GROUP":
        control_node = get_node(node.node_tree, "control")
        density = get_input(control_node, "density").default_value
        nodes.remove(node)
        new_node = nodes.new("GeometryNodeDistributePointsOnFaces")
        new_node.label = name
        new_node.distribute_method = "POISSON"
        new_node.inputs["Density Max"].default_value = density
        new_node.inputs["Seed"].default_value = int(name)
    else:
        nodes.remove(node)
        new_node = nodes.new("GeometryNodeGroup")
        new_node.node_tree = node_group
        get_input(new_node, "seed").default_value = int(name)

    new_node.name = name
    new_node.hide = True
    links.new(input_socket, new_node.inputs[0])
    links.new(new_node.outputs[0], output_socket)
    organise_list_group(mod_group)

    return new_node


def create_biome_json(self, context, biomes):
    data = {}

    # ADDED IN ADDON VERSION 2.1.5, SAVED_BIOME_VERSION 1.1
    data["addon_version"] = ADDON_VERSION

    data["version"] = SAVED_BIOME_VERSION
    j_biomes = {}

    def add_node_inputs(node, dict):
        for socket in node.inputs:
            if socket.type not in ["GEOMETRY", "OBJECT", "COLLECTION"] and not socket.links:
                if socket.type == "RGBA":
                    val = socket.default_value
                    dict[socket.name] = (val[0], val[1], val[2], val[3])
                else:
                    dict[socket.name] = socket.default_value
        return dict

    for biome in biomes:
        if not biome.biome_settings:
            continue
        j_biome = {}
        j_biome["use_scene_cam"] = biome.biome_settings.at.use_scene_cam
        bsettings = biome.biome_settings

        j_biome["node_settings"] = {}
        geo_control_node = get_node(bsettings, "control")
        add_node_inputs(geo_control_node, j_biome["node_settings"])

        j_biome["trees"] = {}
        tree_list = bsettings.at.tree_list
        for tree in tree_list:
            j_tree = {}
            j_tree["selected_tree"] = tree.selected_tree
            j_tree["quality"] = tree.quality

            j_tree["node_settings"] = {}
            tree_control_node = get_node(get_mod_group(tree.tree_obj), "tree")
            add_node_inputs(tree_control_node, j_tree["node_settings"])

            j_mat = {}
            mat_control_node = get_mat_control_node(self, context, tree)
            add_node_inputs(mat_control_node, j_mat)

            j_tree["material_settings"] = j_mat
            j_biome["trees"][tree.name] = j_tree

        j_biomes[biome.biome_settings.at.name] = j_biome

    data["biomes"] = j_biomes

    return data


def load_biome_json(self, context, data):
    not_found = []

    def apply_node_settings(node, data: dict):
        for name, value in data.items():
            try:
                socket = node.inputs[name]
            except KeyError:
                not_found.append(name)
                continue
            socket.default_value = value

    for name, j_biome in data["biomes"].items():
        biome = get_biome_vars(self, context)[-1]
        if not biome or biome.biome_settings:
            bpy.ops.alpha_tree.biome_list_actions(action='ADD')
            biome = get_biome_vars(self, context)[-1]

        bpy.ops.alpha_tree.sys_new_settings(add_tree=False)
        bsettings = biome.biome_settings
        bsettings.at.name = name
        biome.use_scene_cam = j_biome["use_scene_cam"]

        geo_control_node = get_node(bsettings, "control")
        apply_node_settings(geo_control_node, j_biome["node_settings"])

        tree_list = bsettings.at.tree_list
        for name, j_tree in j_biome["trees"].items():
            bpy.ops.alpha_tree.tree_list_actions(action='ADD')
            tree = tree_list[-1]
            tree.name = name
            tree.selected_tree = j_tree["selected_tree"]
            tree.quality = j_tree["quality"]

            tree_control_node = get_node(get_mod_group(tree.tree_obj), "tree")
            apply_node_settings(tree_control_node, j_tree["node_settings"])

            mat_control_node = get_mat_control_node(self, context, tree)
            apply_node_settings(mat_control_node, j_tree["material_settings"])

    return not_found