import contextlib
import os
from statistics import mean

import bpy

from ..constants import AT_LAYER_NAMES, NODE_NAMES, VGROUP_PRESETS, WL_LAYER_NAMES
from ..icons import get_icon, selected_tree_icon_update
from ..weight_layers_addon.WL_constants import GROUP_TYPES
from ..weight_layers_addon.WL_functions import (
    get_node_by_type,
    get_node_tree_inputs,
    get_node_tree_outputs,
    get_wl_mod_group,
)
from ..weight_layers_addon.WL_functions import get_wl_obj as get_at_obj
from ..weight_layers_addon.WL_functions import setup_new_geo_node_tree
from . import biome_functions
from . import helper_functions as helpers
from .helper_functions import (
    get_all_metadata,
    get_biome_slot_mod_keys,
    get_default_obj,
    get_input,
    get_mod_group,
    get_node,
    get_prefs,
)
from .imp_functions import import_tree, switch_images

# Object


def get_obj_area(obj: bpy.types.Object, use_modifiers: bool = False) -> float:
    """Gets the total area of the current object

    Args:
        object (bpy.types.Object): The object to get the area of
        use_modifiers (bool, optional): Whether to evaluate modifiers (slower). Defaults to False.
    """
    if use_modifiers:
        depsgraph = bpy.context.evaluated_depsgraph_get()
        evaluated_obj = obj.evaluated_get(depsgraph)
        obj = evaluated_obj.to_mesh(preserve_all_data_layers=True, depsgraph=depsgraph)
    mesh = obj.data
    return sum(p.area for p in mesh.polygons)


# Importer


def single_selected_tree_update(self, context):
    selected_tree_icon_update(self)


# Biomes


def new_biome_slot(self, context):
    """Add a new biome slot to this object"""
    obj = self.id_data
    list_settings = obj.at
    biome_list = obj.at.biome_list
    mod_group = helpers.get_mod_group(obj)
    final_index = len(obj.at.biome_list) - 1
    if mod_group:
        nodes = mod_group.nodes
        links = mod_group.links
        join_node = nodes[NODE_NAMES["join"]]
    else:
        node_modifier = obj.modifiers.new("AT_GEO_NODES", type="NODES")
        mod_group = node_modifier.node_group
        # in 3.2 adding a modifier doesn't automatically add a node group as well
        if not mod_group:
            mod_group = setup_new_geo_node_tree("hahah")
            node_modifier.node_group = mod_group
        mod_group.name = f"AT_BIOME_LIST_{obj.name}"
        nodes = mod_group.nodes
        links = mod_group.links

        join_node = nodes.new("GeometryNodeJoinGeometry")
        join_node.name = NODE_NAMES["join"]

        output_node = get_node_by_type(mod_group, "GROUP_OUTPUT")
        to_socket = output_node.inputs["Geometry"]
        from_socket = to_socket.links[0].from_socket

        links.new(from_socket, join_node.inputs[0])
        links.new(join_node.outputs[0], to_socket)

    input_node = get_node_by_type(mod_group, "GROUP_INPUT")
    if not input_node:
        raise ValueError(str([n.name for n in nodes]))
    point_dist_node = nodes.new("GeometryNodeDistributePointsOnFaces")
    point_dist_node.name = str(final_index + 1)
    point_dist_node.distribute_method = "POISSON"
    point_dist_node.inputs["Seed"].default_value = final_index + 1
    point_dist_node.hide = True
    links.new(input_node.outputs[0], point_dist_node.inputs[0])
    links.new(point_dist_node.outputs[0], join_node.inputs[0])

    if len(biome_list) > 0:
        last_slot_node = mod_group.nodes.get(str(len(biome_list) - 1))
        if last_slot_node.type == "DISTRIBUTE_POINTS_ON_FACES":
            old_dens_input = last_slot_node.inputs["Density Max"]
            scale = float(last_slot_node.label)
        else:
            old_control_node = get_node(last_slot_node.node_tree, "control")
            old_dens_input = get_input(old_control_node, "density")
            scale = get_input(old_control_node, "scale").default_value
        density = old_dens_input.default_value
    else:
        area = obj.at.get_mesh_area()
        density = 1000 / area
        scale = mean(obj.dimensions) / 5  # approximation
    point_dist_node.inputs["Density Max"].default_value = density
    point_dist_node.label = str(scale)  # Store scale in label as a temprorary property. It's not great but it works...

    biome_functions.organise_list_group(mod_group)

    biome_list.add()
    final_index = len(biome_list) - 1
    new_biome = biome_list[final_index]
    new_biome.name = f"Biome {str(final_index)}"
    list_settings.b_index = final_index

    return new_biome


def biome_index_update(self, context):
    mod_group = helpers.get_mod_group(get_at_obj(context))
    if not mod_group:
        return
    nodes = mod_group.nodes
    for node in nodes:
        if node.type in {"DISTRIBUTE_POINTS_ON_FACES", "GROUP"}:
            if int(node.name) == self.b_index:
                nodes.active = node
                node.select = True
            else:
                node.select = False


def get_biome_settings_enum(self, context):
    """Return enum of all biome settings"""
    items = [("NONE", "None", "No biome settings selected")]
    for group in bpy.data.node_groups:
        if group.type == "GEOMETRY" and group.at.type == "SETTINGS":
            item = (group.name, group.at.name, group.name)
            items.append(item)
    return items


def biome_settings_enum_update(self, context):
    if self.biome_settings_enum == "NONE":
        if self.biome_settings:
            self.biome_settings = None
        return
    if self.biome_settings and self.biome_settings.name == self.biome_settings_enum:
        return
    self.biome_settings = bpy.data.node_groups[self.biome_settings_enum]


def biome_show_viewport_update(self, context):
    mod_group = helpers.get_mod_group(get_at_obj(context))
    index = helpers.get_list_index(self, type="str")
    node = mod_group.nodes[index]
    get_input(node, "show viewport").default_value = self.show_viewport


def biome_show_render_update(self, context):
    mod_group = helpers.get_mod_group(get_at_obj(context))
    index = helpers.get_list_index(self, type="str")
    node = mod_group.nodes[index]
    get_input(node, "show render").default_value = self.show_render


def biome_settings_update(self, context):
    obj = self.id_data
    modifier = helpers.get_mod_group(obj, modifier=True)
    node_group = modifier.node_group
    nodes = node_group.nodes
    b_index = obj.at.b_index
    group_input = get_node_by_type(node_group, "GROUP_INPUT")
    prev_node = nodes.get(str(b_index))
    sockets = [get_input(prev_node, "density_attr"), get_input(prev_node, "scale_attr")]

    def add_modifier_inputs(node):
        node_group.interface.new_socket(
            name=f"density_input_{str(b_index)}", in_out="INPUT", socket_type="NodeSocketFloat"
        )
        node_group.interface.new_socket(
            name=f"scale_input_{str(b_index)}", in_out="INPUT", socket_type="NodeSocketFloat"
        )
        node_group.links.new(group_input.outputs[-3], get_input(node, "density_attr"))
        node_group.links.new(group_input.outputs[-2], get_input(node, "scale_attr"))
        for k in modifier.keys():
            if "use_attribute" in k and not modifier[k]:
                modifier[k.replace("_use_attribute", "")] = 1.0

    if self.biome_settings:
        if prev_node.type == "GROUP":
            self.biome_settings_enum = self.biome_settings.name
            prev_node.node_tree = self.biome_settings
            return
        else:  # swap nodes
            node = biome_functions.swap_nodes(prev_node, node_group=self.biome_settings)
            add_modifier_inputs(node)

    else:
        self.biome_settings_enum = "NONE"
        for socket in sockets:
            idx = list(get_node_tree_outputs(node_group)).index(socket.links[0].from_socket)
            node_group.interface.remove(get_node_tree_inputs(node_group)[idx])
        biome_functions.swap_nodes(prev_node)

    return


def get_vgroup_enum_items(self, context):
    items = [("NONE", "None", "None")]
    obj = get_at_obj(context)
    items.extend((vg.name, vg.name, vg.name) for vg in obj.vertex_groups)

    # put in alphabetical order, with none at the top
    items.sort(key=lambda item: item[0] if item[0] != "NONE" else "0000000")
    return items


def vgroup_enum_update(self, context):
    if self.vgroup_enum == "NONE":
        self.vgroup = None
        return

    # if vgroup has been removed
    try:
        _ = self.vgroup
    except AttributeError:
        self.vgroup = None

    if self.vgroup and self.vgroup_enum == self.vgroup.name:
        return

    obj = get_at_obj(context)
    self.vgroup = obj.vertex_groups[self.vgroup_enum]


def vgroup_update(self, context):
    """Sets the attribute inputs of the slot node group to the current vgroup name"""
    obj = self.id_data
    wl = obj.weight_layers
    modifier = get_mod_group(obj, modifier=True)
    settings_group = self.biome_slot.biome_settings
    keys = get_biome_slot_mod_keys(settings_group, modifier)
    key = keys[0 if self.density else 1]

    if self.vgroup:
        # update modifier inputs
        modifier[f"{key}_use_attribute"] = True
        modifier[f"{key}_attribute_name"] = self.vgroup.name

        # Check if slot exists, and if not add one with a unique name, which is used to reference it
        slot = None
        with contextlib.suppress(KeyError):
            slot = wl.layer_stacks[self.stack_slot_name]
        if not slot or not self.stack_slot_name:
            slot = wl.new_layer_stack_slot()
            unique_name = ""
            i = 0
            while True:
                dist_type = " Density " if self.density else " Scale "
                name = self.biome_slot.biome_settings.name
                unique_name = name.replace(NODE_NAMES["slot"] + "_", "") + dist_type + str(i)

                try:
                    wl.layer_stacks[unique_name]
                except KeyError:
                    break
                i += 1
            slot.name = unique_name
            self.stack_slot_name = slot.name
        unique_name = slot.name

        # get the layer stack, create if doesn't exist
        stack = None
        for ng in bpy.data.node_groups:
            if ng.type == "GEOMETRY":
                try:
                    wl_vgroup = ng.wl.vgroup
                except AttributeError:
                    wl_vgroup = self.vgroup
                if wl_vgroup == self.vgroup:
                    stack = ng.wl
                    if slot.stack_group != ng:
                        slot.stack_group = ng
                    break
        if not stack:
            stack = slot.new_layer_stack(context, self.vgroup.name, slot)

        # set vgroup in stack
        stack.vgroup = self.vgroup
        for layer in reversed(list(stack.layers)):
            if hasattr(layer.layer_settings, "vgroup_enum") and layer.layer_settings.is_output:
                layer.layer_settings.slot_idx = slot.index
                layer.layer_settings.vgroup_enum = self.vgroup.name
                layer.layer_settings.slot_idx = -1
                break

        input_layer = None
        for layer in list(stack.layers):
            if (
                hasattr(layer.layer_settings, "vgroup_enum")
                and not layer.layer_settings.is_output
                and layer.layer_settings.vgroup_enum == self.vgroup.name
            ):
                input_layer = layer

        if not input_layer:
            input_layer = stack.new_layer(context, WL_LAYER_NAMES["input"])
        input_layer.layer_settings.vgroup_enum = self.vgroup.name

        done = []
        wl_mod_group = get_wl_mod_group(context, obj)
        for i, slot in enumerate(list(wl.layer_stacks)[::-1]):
            if slot.stack_group and slot.stack_group.wl.vgroup:
                node = wl_mod_group.nodes[str(len(wl.layer_stacks) - i - 1)]
                if slot.stack_group in done:
                    node.mute = True
                else:
                    node.mute = False
                    done.append(slot.stack_group)

    else:
        modifier[f"{key}_use_attribute"] = False

        # remove weight layers stack slot
        if wl.layer_stacks:
            for i, slot in enumerate(wl.layer_stacks):
                if slot.name == self.stack_slot_name:
                    for layer in slot.stack_group.wl.layers:
                        if hasattr(layer.layer_settings, "slot_idx"):
                            layer.layer_settings.slot_idx = i
                    break

            wl.remove_layer_stack_slot(context, i)


def vgroup_name_get(self):
    try:
        vgroup = self.vgroup
    except AttributeError:
        return ""
    return vgroup.name if vgroup else ""


def vgroup_name_set(self, value):
    self.vgroup.name = value


def new_vgroup(self, context: bpy.types.Context, type="blank") -> bpy.types.VertexGroup:
    """Adds a new vertex group that is routed through the weight layers addon automatically"""
    obj = self.id_data
    NAME = "AT VGroup"
    vgroup = obj.vertex_groups.new(name=NAME)
    self.vgroup_enum = vgroup.name
    slot = get_at_obj(context).weight_layers.layer_stacks[self.stack_slot_name]
    stack = slot.stack_group.wl

    if type != "blank":
        for layer in VGROUP_PRESETS[type]:
            stack.new_layer(context, layer["type"])

    # Set layer names so they wont be drawn in the Alpha Trees UI
    stack.layers[0].name = AT_LAYER_NAMES["input"]
    stack.layers[-1].name = AT_LAYER_NAMES["output"]

    obj.vertex_groups.active = vgroup
    return vgroup


def remove_vgroup(self, context: bpy.types.Context) -> None:
    """Removes the currently selected vgroup"""
    obj = self.id_data
    vgroup = self.vgroup

    if not vgroup:
        return

    name = vgroup.name
    other_dist = getattr(obj.at.biome_list[obj.at.b_index], "scale" if self.density else "density")
    if other_dist.vgroup_enum == name:
        other_dist.vgroup_enum = "NONE"
    for ng in bpy.data.node_groups:
        if ng.type == "GEOMETRY" and ng.wl.type == GROUP_TYPES["stack"] and ng.wl.vgroup == vgroup:
            ng.wl.vgroup = None

    self.vgroup = None
    obj.vertex_groups.remove(vgroup)


def wlayer_slot_get(self):
    try:
        stack_group = self["wl_stack_group"]
    except KeyError:
        return None
    return stack_group.wl if stack_group else None


def wlayer_slot_set(self, val):
    self["wl_stack_group"] = val.id_data


# Biome settings


def settings_name_update(self, context):
    """Makes sure name is unique and also update collection name"""
    group = self.id_data
    if NODE_NAMES["slot"] + "_" + self.name == group.name:
        return

    index = 0
    # get unique name
    while True:
        try:
            name = NODE_NAMES["slot"] + "_" + self.name + " " + str(index)
            if not index:
                name = name[:-2]
            _ = bpy.data.node_groups[name]
            index += 1
        except KeyError:
            break

    group.name = name
    control_node = get_node(group, "control")
    if collection := get_input(control_node, "collection").default_value:
        collection.name = name
        self.name = name[len(NODE_NAMES["slot"] + "_") :]


# Trees


def tree_show_render_update(self, context):
    mod_group = helpers.get_mod_group(self.tree_obj)
    node = get_node(mod_group, "tree")
    get_input(node, "tree_show_render").default_value = self.show_render
    return


def get_tree_icon_value(self):
    # sourcery skip: assign-if-exp, inline-immediately-returned-variable
    """Return the integer value of the icon for the given tree"""

    if get_prefs().use_mat_icons:
        return self.tree_obj.active_material.preview.icon_id
    else:
        return get_icon(self.selected_tree, type="preview")


def get_material_icon_value(self):
    """Return the integer value of the alpha trees icon for the given tree material"""
    mat = self.id_data
    if get_prefs().use_mat_icons:
        return mat.preview.icon_id
    if diff_node := get_node(mat.node_tree, "leaf_col"):
        prev_name = diff_node.image.name.replace("diff-leaf.jpg", "preview.png")
    return get_icon(prev_name, type="preview")


def selected_tree_update(self, context):
    settings_group = self.id_data

    at = context.scene.alpha_trees
    orig_object = context.object
    selected = orig_object.select_get() if orig_object else False

    control_node = get_node(settings_group, "control")
    collection = get_input(control_node, "collection").default_value

    if self.selected_tree == self.previous_tree and self.selected_tree != "default.png":
        return

    # get previous node tree
    if self.tree_obj and self.previous_tree != "default.png":
        mods = self.tree_obj.modifiers
        geo_mods = [m for m in mods if m.type == "NODES"]
        mod_group = geo_mods[0].node_group
    else:
        mod_group = None

    if self.selected_tree == "default.png":
        object = get_default_obj(self, context, control_node)
    else:
        # check that LODs exist
        path = os.path.join(
            get_prefs().lib_path,
            "rendered_maps",
            self.quality if self.quality != "high" else "",
            self.selected_tree.replace("preview.png", "diff-leaf.jpg"),
        )
        self.quality = "high" if not os.path.isfile(path) else self.quality

        import_tree(os.path.join(path, "rendered_maps", self.selected_tree), False, self.quality)
        bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)

        object = context.active_object
        object.location = (0, 0, 0)

        tree_name = self.selected_tree.split("_")[0] + "_"

        # remove previous tree
        if self.previous_tree == "default.png":
            for obj in collection.objects:
                if "AT_default_obj" in obj.name:
                    previous_tree = obj
                    break
        else:
            previous_tree = self.tree_obj

        collection.objects.unlink(previous_tree)
        try:
            bpy.data.materials.remove(previous_tree.material_slots[0].material)
        except (IndexError, AttributeError):
            pass
        bpy.data.meshes.remove(previous_tree.data)

        # get object name and avoid duplicates
        obj_index = 0
        while True:
            try:
                _ = bpy.data.objects["AT_PARTICLE_" + tree_name + str(obj_index)]
                obj_index += 1
            except KeyError:
                break

        self.name = tree_name + str(obj_index)
        object.name = "AT_PARTICLE_" + tree_name + str(obj_index)

    self.tree_obj = object

    collection.objects.link(object)

    # check that original object exists (needed because objects can be pinned)
    if orig_object:
        try:
            orig_object.select_set(selected)
            bpy.context.view_layer.objects.active = orig_object
        except (RuntimeError, ReferenceError):
            pass

    for obj in context.scene.collection.objects:
        if obj.name == object.name:
            context.scene.collection.objects.unlink(object)
            break
    else:
        for coll in bpy.data.collections:
            if NODE_NAMES["slot"] not in coll.name:
                for obj in coll.objects:
                    if obj.name == object.name:
                        coll.objects.unlink(object)
                        break

    if not helpers.get_mod_group(self.tree_obj):
        node_modifier = self.tree_obj.modifiers.new("AT_GEO_NODE_TREE", type="NODES")
        if node_modifier.node_group:
            mod_group = node_modifier.node_group
        else:
            mod_group = node_modifier.node_group
            if not mod_group:
                mod_group = setup_new_geo_node_tree("AT_GEO_NODE_TREE_" + self.name)
                node_modifier.node_group = mod_group
        #     bpy.data.node_groups.remove(node_modifier.node_group)
        # else:
        biome_functions.setup_tree_group(mod_group)
        node_modifier.node_group = mod_group
        mod_group.name = "AT_GEO_NODE_TREE_" + self.name

    if "Biome settings" in settings_group.at.name:
        metadata = get_all_metadata()[self.selected_tree.replace("_preview.png", "")]
        settings_group.at.name = metadata["subtype"]

    self.previous_tree = self.selected_tree
    at.update_selected_tree = True
    self.scale = 1

    selected_tree_icon_update(self)  # set this tree to be loaded at high res


def quality_update(self, context):
    if isinstance(self.id_data, bpy.types.Material):
        mat = self.id_data
        nodes = mat.node_tree.nodes
        diff_node = get_node(mat.node_tree, "leaf_col")
        if diff_node.image:
            name = diff_node.image.name
            name = name[: name.index(".jpg") + 4]
            name = name.replace("diff-leaf.jpg", "preview.png")
        else:
            return

    else:
        if "default.png" in self.selected_tree:
            return
        name = self.selected_tree
        obj = self.tree_obj
        if not obj.active_material:
            return
        nodes = obj.active_material.node_tree.nodes

    path = os.path.join(
        get_prefs().lib_path,
        "rendered_maps",
        self.quality if self.quality != "high" else "",
        name.replace("preview.png", "diff-leaf.jpg"),
    )
    if not os.path.isfile(path):
        return

    quality = self.quality

    images = helpers.get_other_maps(context, name, quality)
    switch_images(images, nodes)

    return


# Copy items


def copy_item_on_update(self, context):
    if self.exclusive and self.on:
        for i in self.id_data.alpha_trees.copy_items:
            if i != self:
                i.on = False
