import bpy

import os
from os.path import join as osjoin
import re

from . import helper_functions as helpers
from .helper_functions import get_prefs
from ..icons import get_icon
from ..constants import AT_CAM_NAME, AT_WORLD_NAME, AT_PREV_WORLD_NAME, AT_BLEND_NAME

# UTILS
# ----------------------------------------------------------------------------


def tree_obj_poll(self, object):

    if object.type != "MESH" and object.type != "EMPTY":
        return False

    for item in bpy.context.scene.alpha_trees.obj_list:
        if item == self:
            continue
        elif object == item.tree_obj:
            return False

    for obj in bpy.context.scene.objects:
        if obj == object:
            return True
    else:
        return False


def tree_obj_update(self, context):
    at = bpy.context.scene.alpha_trees
    if self.tree_obj and at.auto_material_update:
        bpy.ops.alpha_tree.auto_mat_type()  # pylint: disable=no-member


def mat_type_items(self, context):
    items = (
        ("TRUNK", "Trunk", "Material is used for the trunk", get_icon("dead_tree.png"), 0),
        ("BRANCH", "Branch", "Material is used for the branches", get_icon("branch.png"), 1),
        ("LEAF", "Leaf", "Material is used for the leaves", get_icon("leaf.png"), 2),
        ("FRUIT", "Fruit", "Material is used for the fruit", get_icon("apple.png"), 3),
        ("FLOWER", "Flower", "Material is used for the flower", get_icon("flower.png"), 4),
    )

    return items


def mat_type_update(self, context):
    self.unsure = False


def rename_output(path, format=".png"):
    # vars
    frame = str(bpy.context.scene.frame_current)
    zero_length = 4 - len(frame)

    full_frame = ""
    for _ in range(zero_length):
        full_frame += "0"
    full_frame += str(frame)

    if path.endswith(format):
        path = path[:-len(format)]

    if os.path.isfile(path + format):
        os.remove(path + format)

    os.rename(path + full_frame + format, path + format)


def check_botaniq_ob(object_name):
    name_up = object_name.upper()
    prefixes = ["TREE_", "SHRUB_", "FLOWER_", "PLANT_", "WEED_", "GRASS_", "SUCCULENT_", ""]
    letters = ["_A_", "_B_", "_C_", "_D_", "_E_", "_F_"]
    seasons = ["SPRING", "SUMMER", "AUTUMN", "WINTER"]
    if any(x in name_up for x in prefixes) and any(x in name_up for x in letters)\
            and any(x in name_up for x in seasons):
        try:
            tree_name = object_name.split("_")
            _ = tree_name[1]
            return True
        except IndexError:
            return False
    else:
        return False


def make_tree_editable(data):
    object = data.tree_obj
    if object.type == "EMPTY":
        empty = object
        empty_name = object.name
        try:
            bpy.ops.material.biq_make_selection_editable()  # pylint: disable=no-member
        except AttributeError:
            bpy.ops.object.duplicates_make_real(use_base_parent=True)

        bpy.ops.object.make_local(type='ALL')

        # remove empty
        name_regex = re.compile(r".\d\d\d")
        mo = name_regex.search(empty_name)
        if mo:
            if mo.group():
                for object in bpy.context.scene.objects:
                    if empty_name[:-3] in object.name and empty_name != object.name:
                        tree = object
                        bpy.data.objects.remove(empty)
                        break
        else:
            bpy.data.objects.remove(empty)
            tree = bpy.data.objects[empty_name + ".001"]
        tree.name = empty_name
        tree.select_set(True)
        bpy.context.view_layer.objects.active = tree
    else:
        tree = object

    # stop the material types being auto picked when generating a tree and losing user input
    at = bpy.context.scene.alpha_trees
    at.auto_material_update = False
    data.tree_obj = tree
    at.auto_material_update = True
    return (tree)


def set_render_border(context, data):
    at = bpy.context.scene.alpha_trees
    object = data.tree_obj
    at = context.scene.alpha_trees
    context.scene.render.use_border = True
    object = make_tree_editable(data)

    size_x = object.dimensions[0]
    size_y = object.dimensions[2]

    context.scene.render.border_min_x = 0.5 - size_x/20 - at.border_padding
    context.scene.render.border_min_y = 0
    context.scene.render.border_max_x = 0.5 + size_x/20 + at.border_padding
    context.scene.render.border_max_y = size_y/10 + at.border_padding / 2


def separate_materials(self, context, data):
    # separate object
    object = data.tree_obj
    mat_slots = object.material_slots

    bpy.ops.object.select_all(action='DESELECT')
    object.select_set(True)
    bpy.ops.object.mode_set(mode="EDIT")
    bpy.ops.mesh.reveal()
    bpy.ops.mesh.select_all(action='DESELECT')

    for i, slot in enumerate(mat_slots):
        if any(type in slot.material.at.type for type in ["FRUIT", "LEAF", "FLOWER"]):
            object.active_material_index = i
            bpy.ops.object.material_slot_select()

    try:
        bpy.ops.mesh.separate(type='SELECTED')
    except RuntimeError as er:
        if "NOTHING SELECTED" in str(er).upper():
            self.report(
                {"ERROR"}, "The leaf, fruit or flower materials are not applied to any geometry, cant split mesh"
            )
        else:
            self.report({"ERROR"}, "Could not split mesh, please check that your materials are set up properly")
        bpy.data.scenes.remove(bpy.data.scenes["alpha_trees_scene"])
        bpy.ops.object.mode_set(mode="OBJECT")
        return None, None

    bpy.ops.object.mode_set(mode="OBJECT")
    bark_object = object
    for obj in context.selected_objects:
        if obj != bark_object:
            leaf_object = obj
    return bark_object, leaf_object


def unlink_file_output_nodes(node_tree):
    bpy.context.scene.use_nodes = True
    for node in node_tree.nodes:
        if node.type == "OUTPUT_FILE":
            for input in node.inputs:
                if input.links:
                    node_tree.links.remove(input.links[0])


def unlink_output_node(compositor=True, material=None):
    # unlink composite node
    output_link = None
    composite_link = None
    if compositor:
        node_tree = bpy.context.scene.node_tree
        for node in node_tree.nodes:
            if node.type == "COMPOSITE":
                if node.inputs[0].links:
                    output = node.inputs[0].links[0]
                    output_link = output.from_socket
                    composite_link = output.to_socket
                    node_tree.links.remove(output)
    else:
        mat_output_node = material.node_tree.nodes.new("ShaderNodeOutputMaterial")
        for node in material.node_tree.nodes:
            if node.type == 'OUTPUT_MATERIAL':
                node.is_active_output = False
        mat_output_node.is_active_output = True

    if compositor:
        return output_link, composite_link
    else:
        return mat_output_node


def fast_render(use_scene_res=False):
    render = bpy.context.scene.render
    x_res = render.resolution_x
    y_res = render.resolution_y
    orig_settings = [x_res, y_res, render.engine]
    render.resolution_x = x_res if use_scene_res else 10
    render.resolution_y = y_res if use_scene_res else 10
    render.engine = "BLENDER_WORKBENCH"

    bpy.ops.render.render(use_viewport=True)

    render.resolution_x = orig_settings[0]
    render.resolution_y = orig_settings[1]
    render.engine = orig_settings[2]


def get_settings():
    render = bpy.context.scene.render
    view_settings = bpy.context.scene.view_settings
    display = bpy.context.scene.display
    shading = display.shading
    cycles = bpy.context.scene.cycles

    settings = [
        bpy.context.scene.use_nodes,

        # render settings
        render.engine,
        render.filepath,
        render.film_transparent,
        render.resolution_percentage,
        render.resolution_x,
        render.resolution_y,
        render.use_border,
        render.use_crop_to_border,
        render.use_file_extension,
        render.use_overwrite,
        render.use_stamp,
        render.pixel_aspect_x,
        render.pixel_aspect_y,
        render.use_single_layer,
        render.use_compositing,
        render.use_sequencer,
        render.image_settings.file_format,
        render.image_settings.quality,
        cycles.transparent_max_bounces,

        # color management settings
        view_settings.exposure,
        view_settings.gamma,
        view_settings.look,
        view_settings.use_curve_mapping,
        view_settings.view_transform,
        bpy.context.scene.display_settings.display_device,
        bpy.context.scene.sequencer_colorspace_settings.name,

        # workbench settings
        display.render_aa,
        shading.show_cavity,
        shading.show_backface_culling,
        shading.show_object_outline,
        shading.show_shadows,
        shading.show_specular_highlight,
        shading.show_xray,
        shading.use_dof,
        shading.light,
        shading.studio_light,
        shading.color_type,
    ]

    return settings


def set_general_settings():
    render = bpy.context.scene.render
    view_settings = bpy.context.scene.view_settings
    display = bpy.context.scene.display
    shading = display.shading
    at = bpy.context.scene.alpha_trees
    cycles = bpy.context.scene.cycles

    bpy.context.scene.use_nodes = True

    # render settings
    render.film_transparent = True
    render.resolution_percentage = 100
    render.resolution_x = at.resolution
    render.resolution_y = at.resolution
    render.use_border = True
    render.use_crop_to_border = True
    render.use_file_extension = True
    render.use_overwrite = at.overwrite
    render.use_stamp = False
    render.pixel_aspect_x = 1
    render.pixel_aspect_y = 1
    render.use_single_layer = True
    render.use_compositing = True
    render.use_sequencer = False
    render.image_settings.file_format = "JPEG"
    render.image_settings.quality = 85

    cycles.transparent_max_bounces = 30

    # color management settings
    view_settings.exposure = 0
    view_settings.gamma = 1
    view_settings.look = "None"
    view_settings.use_curve_mapping = False
    view_settings.view_transform = "Standard"
    bpy.context.scene.display_settings.display_device = "sRGB"
    bpy.context.scene.sequencer_colorspace_settings.name = "Filmic Log"

    # workbench settings
    display.render_aa = "32"
    shading.show_cavity = False
    shading.show_backface_culling = False
    shading.show_object_outline = False
    shading.show_shadows = False
    shading.show_specular_highlight = False
    shading.show_xray = False
    shading.use_dof = False


def reset_settings(settings):
    # This is a terrible system that I'm going to replace soon. How I ever thought it was a good idea is beyond me...
    render = bpy.context.scene.render
    view_settings = bpy.context.scene.view_settings
    display = bpy.context.scene.display
    shading = display.shading
    cycles = bpy.context.scene.cycles

    i = 0

    bpy.context.scene.use_nodes = settings[i]
    i += 1

    # render settings
    render.engine = settings[i]
    i += 1
    render.filepath = settings[i]
    i += 1
    render.film_transparent = settings[i]
    i += 1
    render.resolution_percentage = settings[i]
    i += 1
    render.resolution_x = settings[i]
    i += 1
    render.resolution_y = settings[i]
    i += 1
    render.use_border = settings[i]
    i += 1
    render.use_crop_to_border = settings[i]
    i += 1
    render.use_file_extension = settings[i]
    i += 1
    render.use_overwrite = settings[i]
    i += 1
    render.use_stamp = settings[i]
    i += 1
    render.pixel_aspect_x = settings[i]
    i += 1
    render.pixel_aspect_y = settings[i]
    i += 1
    render.use_single_layer = settings[i]
    i += 1
    render.use_compositing = settings[i]
    i += 1
    render.use_sequencer = settings[i]
    i += 1
    render.image_settings.file_format = settings[i]
    i += 1
    render.image_settings.quality = settings[i]
    i += 1

    i += 1
    cycles.transparent_max_bounces = int(settings[i])

    # color management settings
    view_settings.exposure = settings[i]
    i += 1
    view_settings.gamma = settings[i]
    i += 1
    view_settings.look = settings[i]
    i += 1
    view_settings.use_curve_mapping = settings[i]
    i += 1
    view_settings.view_transform = settings[i]
    i += 1
    bpy.context.scene.display_settings.display_device = settings[i]
    i += 1
    bpy.context.scene.sequencer_colorspace_settings.name = settings[i]
    i += 1

    # workbench settings
    display.render_aa = settings[i]
    i += 1
    shading.show_cavity = settings[i]
    i += 1
    shading.show_backface_culling = settings[i]
    i += 1
    shading.show_object_outline = settings[i]
    i += 1
    shading.show_shadows = settings[i]
    i += 1
    shading.show_specular_highlight = settings[i]
    i += 1
    shading.show_xray = settings[i]
    i += 1
    shading.use_dof = settings[i]
    i += 1
    shading.light = settings[i]
    i += 1
    shading.studio_light = settings[i]
    i += 1
    shading.color_type = settings[i]
    i += 1


def load_image(file_name, img_path, name, format=".jpg"):
    # load
    for image in bpy.data.images:
        if image.name == name + file_name + format:
            bpy.data.images.remove(image)
    image = bpy.data.images.load(img_path + file_name + format)
    return image


# after generation


def generate_lods(context, tree_name):
    # debug
    if tree_name == "default.png":
        tree_name = "Beech 1"

    high_path = bpy.path.abspath(osjoin(get_prefs().lib_path, "rendered_maps"))
    images = [f for f in os.listdir(high_path) if tree_name + "_" in f and "_preview" not in f and ".json" not in f]
    med_path = osjoin(high_path, "med")
    low_path = osjoin(high_path, "low")

    # setup new scene
    orig_scene = context.scene
    bpy.ops.scene.new(type='NEW')
    scene = context.scene
    bpy.ops.object.camera_add()
    cam = context.object
    scene.camera = cam
    scene.view_settings.view_transform = 'Standard'
    scene.use_nodes = True
    scale_nodes = []
    b_images = []
    nodes = scene.node_tree.nodes
    links = scene.node_tree.links
    nodes.clear()

    # setup nodes
    out_node = nodes.new("CompositorNodeOutputFile")
    out_node.file_slots.clear()

    rl_node = nodes.new("CompositorNodeRLayers")

    b_img = None
    for img in images:
        out_node.file_slots.new(os.path.splitext(img)[0])
        slot = out_node.file_slots[-1]
        slot.use_node_format = False
        slot.save_as_render = True
        file_type = os.path.splitext(img)[1]
        img_type = "JPEG" if file_type == ".jpg" else "PNG"
        slot.format.file_format = img_type

        img_path = bpy.path.abspath(osjoin(high_path, img))
        for b_img in bpy.data.images:
            if b_img.filepath_raw == img_path:
                break
        else:
            b_img = bpy.data.images.load(img_path)
        b_images.append(b_img)

        img_node = nodes.new("CompositorNodeImage")
        img_node.image = b_img

        scale_node = nodes.new("CompositorNodeScale")
        scale_nodes.append(scale_node)

        mix_node = nodes.new("CompositorNodeMixRGB")

        blur_node = nodes.new("CompositorNodeBlur")
        blur_node.filter_type = "MITCH"
        blur_node.size_x = 2
        blur_node.size_y = 2
        blur_node.inputs[1].default_value = 0.85

        links.new(img_node.outputs[0], scale_node.inputs[0])
        links.new(scale_node.outputs[0], mix_node.inputs[2])
        links.new(rl_node.outputs[0], mix_node.inputs[1])
        links.new(mix_node.outputs[0], blur_node.inputs[0])
        links.new(blur_node.outputs[0], out_node.inputs[-1])

    if not b_img:
        bpy.data.cameras.remove(cam.data)
        bpy.data.scenes.remove(scene)
        context.window.scene = orig_scene
        return

    for i, path in enumerate([med_path, low_path]):

        if not os.path.isdir(path):
            os.mkdir(path)

        # 4, 8
        factor = (i+1) * 4

        out_node.base_path = path
        for node in scale_nodes:
            node.inputs[1].default_value = 1 / factor
            node.inputs[2].default_value = 1 / factor

        context.scene.render.resolution_x = int(b_img.size[0] / factor)
        context.scene.render.resolution_y = int(b_img.size[1] / factor)

        fast_render(use_scene_res=True)

        for img in images:

            rename_output(path=osjoin(path, img), format=os.path.splitext(img)[1])

    bpy.data.cameras.remove(cam.data)
    bpy.data.scenes.remove(scene)
    context.window.scene = orig_scene

    for img in b_images:
        if img.users == 0:
            bpy.data.images.remove(img)

    return


# OPERATORS
# ----------------------------------------------------------------------------------------


def setup_scene(self, context, data):

    for scene in bpy.data.scenes:
        if scene.name == "alpha_trees_scene":
            context.window.scene = scene
            scene = context.scene
            break
    else:
        bpy.ops.scene.new(type='EMPTY')
        scene = context.scene
        scene.name = "alpha_trees_scene"

    at = scene.alpha_trees
    tree_obj = data.tree_obj
    file_path = str(helpers.get_resources_path(sub=AT_BLEND_NAME))

    already_created = []
    with bpy.data.libraries.load(file_path) as (data_from, data_to):
        for object in data_from.objects:
            if "AT_" in object:
                data_to.objects.append(object)

        for object in data_from.objects:
            if "AT_" in object:
                for data_object in bpy.data.objects:
                    if object == data_object.name:
                        data_to.objects.remove(object)
                        already_created.append(data_object)

        for world in bpy.data.worlds:
            if world.name == AT_WORLD_NAME:
                bpy.context.scene.world = world
                break
        else:
            data_to.worlds.append(AT_WORLD_NAME)

    try:
        collection = bpy.context.scene.collection.children["alpha_generator_objects"]
    except KeyError:
        collection = bpy.data.collections.new("alpha_generator_objects")
        bpy.context.scene.collection.children.link(collection)

    for object in already_created:
        try:
            collection.objects.link(object)
        except RuntimeError:
            continue

    for object in data_to.objects:
        if object.type == "CAMERA":
            context.scene.camera = collection.objects.link(object)
        else:
            collection.objects.link(object)

    if data_to.worlds:
        bpy.context.scene.world = data_to.worlds[0]

    if not context.scene.camera:
        context.scene.camera = context.scene.objects[AT_CAM_NAME]

    # for object in context.scene.objects:
    #     if object.type == "CAMERA":
    #         context.scene.camera = object
    #         break
    # else:
    # camera = bpy.data.cameras.new(AT_CAM_NAME)
    # cam_obj = bpy.data.objects.new("camera_object", camera)
    # context.scene.collection.objects.link(cam_obj)
    # context.scene.camera = cam_obj
    # xprint(context.scene.camera.name)

    # try:
    context.scene.collection.objects.link(tree_obj)
    context.view_layer.objects.active = tree_obj
    # except RuntimeError:
    #     pass

    scene.render.resolution_x = at.resolution
    scene.render.resolution_y = at.resolution


def scale_to_height(data):
    object = data.tree_obj
    target = 10
    padding = 0.03
    bpy.context.scene.use_nodes = True

    # catches error if mesh is multi user
    try:
        bpy.ops.object.transform_apply(location=False, rotation=True, scale=False)
    except RuntimeError:
        obj = bpy.context.object
        # name = ob.data.name
        # new_mesh = ob.data.copy()
        # new_mesh.name = name
        # ob.data = new_mesh
        obj.data.user_clear()
        obj.data = obj.data
        bpy.ops.object.transform_apply(location=False, rotation=True, scale=False)

    # if linked
    object = make_tree_editable(data)

    # check if needed
    if round(object.dimensions[2]) == round(target - padding):
        return

    dimensions = object.dimensions
    if dimensions[0] > dimensions[2]:
        z_scale = target / dimensions[0] - padding
    else:
        z_scale = target / dimensions[2] - padding

    object.scale = [z_scale * object.scale[0], z_scale * object.scale[1], z_scale * object.scale[2]]


def render_maps(self, context, data):

    object = data.tree_obj
    ob_name = object.name

    scene = context.scene
    setup_scene(self, context, data)
    set_render_border(context, data)

    settings = get_settings()
    at = bpy.context.scene.alpha_trees

    path = osjoin(bpy.path.abspath(get_prefs().lib_path), "rendered_maps", "")
    at = bpy.context.scene.alpha_trees
    object.location = (0, 0, 0)

    if check_botaniq_ob(ob_name):
        ob_name = object.name.split("_")
        name = ob_name[1] + "-" + ob_name[2] + "_"

    else:
        name = ob_name.replace("_", "-") + "_"

    img_path = osjoin(path, name)

    # set up text
    orig_selected = context.object
    bpy.ops.object.text_add()
    text_obj = context.object
    context.view_layer.objects.active = orig_selected
    text_obj.rotation_euler = (90, 0, 0)
    text_obj.location = (-5, -5, 0)
    text_obj.scale = (0.75, 0.75, 0.75)
    text_obj.color = (1, 0.45, 0.04, 1)
    text_obj.hide_render = True
    text_obj.show_in_front

    comp_nodes = []

    bark_mats = []
    leaf_mats = []
    for s in object.material_slots:
        if s.material.at.type in ["TRUNK", "BRANCH"]:
            bark_mats.append(s.material)
        elif s.material.at.type in ["LEAF", "FRUIT", "FLOWER"]:
            leaf_mats.append(s.material)

    set_general_settings()
    trunk_object, leaf_object = separate_materials(self, context, data)

    if not trunk_object:
        return

    node_tree = bpy.context.scene.node_tree
    unlink_file_output_nodes(node_tree)

    for node in node_tree.nodes:
        if node.type == "R_LAYERS":
            r_layers = node
            break
    else:
        r_layers = node_tree.nodes.new("CompositorNodeRLayers")

    output_link, composite_link = unlink_output_node()

    # file output
    output_node = node_tree.nodes.new("CompositorNodeOutputFile")
    comp_nodes.append(output_node)
    output_node.base_path = path
    output_node.format.file_format = "JPEG"
    output_node.format.quality = 100
    output_node.format.color_mode = "RGB"
    output_node.file_slots.clear()

    output_node.file_slots.new(name + "nor-leaf")  # 0
    output_node.file_slots.new(name + "nor-trunk")  # 1
    output_node.file_slots.new(name + "diff-leaf")  # 2
    output_node.file_slots.new(name + "diff-trunk")  # 3
    output_node.file_slots.new(name + "mask_leaf")  # 4
    output_node.file_slots.new(name + "mask_trunk")  # 5
    output_node.file_slots.new(name + "mask_shadow")  # 6
    output_node.file_slots.new(name + "mask")  # 7
    output_node.file_slots.new(name + "preview")  # 8

    for i in [4, 5, 6, 7]:
        output_node.file_slots[i].use_node_format = False
        output_node.file_slots[i].format.file_format = "PNG"
        output_node.file_slots[i].format.color_depth = "8"
        output_node.file_slots[i].format.color_mode = "RGB"
    output_node.file_slots[7].format.compression = 100

    output_node.file_slots[8].use_node_format = False
    output_node.file_slots[8].format.file_format = "PNG"
    output_node.file_slots[8].format.compression = 15
    output_node.file_slots[8].format.color_depth = "8"
    output_node.file_slots[8].format.color_mode = "RGBA"

    # nor-leaf_exists = os.path.isfile(img_path + "nor-leaf.jpg")
    # nor-trunk_exists = os.path.isfile(img_path + "nor-trunk.jpg")
    leaf_exists = os.path.isfile(img_path + "diff-leaf.jpg")
    trunk_exists = os.path.isfile(img_path + "diff-trunk.jpg")
    mask_exists = os.path.isfile(img_path + "mask.jpg")

    diff_trunk_img = None
    mask_leaf_img = None

    denoise_node = node_tree.nodes.new("CompositorNodeDenoise")
    comp_nodes.append(denoise_node)
    bpy.context.view_layer.cycles.denoising_store_passes = True

    # D I F F U S E
    if at.diff_render or at.nor_render:
        if not leaf_exists and not trunk_exists or at.overwrite:
            print("")
            print("RENDERING DIFFUSE AND NORMAL:")
            print("Leaves...")
            print("")
            render = bpy.context.scene.render
            render.engine = "CYCLES"
            links = node_tree.links

            # denoise node
            if at.mask_render:
                alpha_link = node_tree.links.new(r_layers.outputs["Alpha"], output_node.inputs[4])
            links.new(r_layers.outputs["Image"], denoise_node.inputs[0])
            links.new(r_layers.outputs["Denoising Normal"], denoise_node.inputs[1])
            links.new(r_layers.outputs["Denoising Albedo"], denoise_node.inputs[2])
            if at.diff_render:
                leaf_link = links.new(denoise_node.outputs[0], output_node.inputs[2])

            # normal nodes
            context.view_layer.use_pass_normal = True

            sep_rgb_node = node_tree.nodes.new("CompositorNodeSepRGBA")
            links.new(r_layers.outputs["Normal"], sep_rgb_node.inputs[0])
            comp_nodes.append(sep_rgb_node)

            comb_rgb_node = node_tree.nodes.new("CompositorNodeCombRGBA")
            comb_rgb_node.inputs[2].default_value = 1
            comb_rgb_node.inputs[3].default_value = 1
            links.new(sep_rgb_node.outputs[0], comb_rgb_node.inputs[0])
            links.new(sep_rgb_node.outputs[2], comb_rgb_node.inputs[1])
            comp_nodes.append(comb_rgb_node)

            mult_node = node_tree.nodes.new("CompositorNodeMixRGB")
            mult_node.blend_type = "MULTIPLY"
            mult_node.inputs[0].default_value = 1
            mult_node.inputs[2].default_value = (0.499458, 0.499458, 0.499458, 1)
            links.new(comb_rgb_node.outputs[0], mult_node.inputs[1])
            comp_nodes.append(mult_node)

            add_node = node_tree.nodes.new("CompositorNodeMixRGB")
            add_node.blend_type = "ADD"
            add_node.inputs[0].default_value = 1
            add_node.inputs[2].default_value = (0.499458, 0.499458, 0.499458, 1)
            links.new(mult_node.outputs[0], add_node.inputs[1])
            comp_nodes.append(add_node)

            nor_blur_node = node_tree.nodes.new("CompositorNodeBlur")
            nor_blur_node.filter_type = "GAUSS"
            nor_blur_node.use_relative = True
            nor_blur_node.factor_x = 0.15
            nor_blur_node.factor_y = 0.15
            links.new(add_node.outputs[0], nor_blur_node.inputs[0])
            comp_nodes.append(nor_blur_node)

            background_node = node_tree.nodes.new("CompositorNodeMixRGB")
            background_node.inputs[1].default_value = (0, 0, 0, 1)
            links.new(r_layers.outputs["Alpha"], background_node.inputs[0])
            links.new(nor_blur_node.outputs[0], background_node.inputs[2])
            comp_nodes.append(background_node)

            if at.nor_render:
                nor_link = links.new(background_node.outputs[0], output_node.inputs[0])

            # holdout bark
            shader_nodes = []
            for mat in bark_mats:
                mat_output_node = unlink_output_node(compositor=False, material=mat)
                holdout_node = mat.node_tree.nodes.new("ShaderNodeHoldout")
                mat.node_tree.links.new(holdout_node.outputs[0], mat_output_node.inputs[0])
                shader_nodes.append([holdout_node, mat_output_node])

            area = context.area
            try:
                area.spaces[0].region_3d.view_perspective = 'CAMERA'
            except AttributeError:
                pass
            trunk_object.hide_viewport = True
            text_obj.data.body = "Rendering leaves..."

            bpy.ops.wm.redraw_timer(type='DRAW_WIN_SWAP', iterations=1)

            # leaves
            scene2 = context.scene
            scene2.render.resolution_x = scene2.render.resolution_y = scene2.alpha_trees.resolution
            bpy.ops.render.render()

            trunk_object.hide_viewport = False

            # trunk
            print("")
            print("Trunk...")
            print("")
            if at.diff_render:
                links.remove(leaf_link)
                trunk_link = links.new(denoise_node.outputs[0], output_node.inputs[3])
            if at.nor_render:
                links.remove(nor_link)
                nor_link = links.new(background_node.outputs[0], output_node.inputs[1])
            if at.mask_render:
                links.remove(alpha_link)

            for i, mat in enumerate(bark_mats):
                for node in shader_nodes[i]:
                    mat.node_tree.nodes.remove(node)

            leaf_object.hide_render = True
            leaf_object.hide_viewport = True
            text_obj.data.body = "Rendering trunk..."
            bpy.ops.wm.redraw_timer(type='DRAW_WIN_SWAP', iterations=1)

            bpy.ops.render.render()

            leaf_object.hide_render = False
            leaf_object.hide_viewport = False

            if at.diff_render:
                links.remove(trunk_link)
                rename_output(img_path + "diff-leaf", format=".jpg")
                rename_output(img_path + "diff-trunk", format=".jpg")
            if at.nor_render:
                links.remove(nor_link)
                rename_output(img_path + "nor-leaf", format=".jpg")
                rename_output(img_path + "nor-trunk", format=".jpg")
            if at.mask_render:
                rename_output(img_path + "mask_leaf", format=".png")

            if at.mask_render:
                mask_leaf_img = load_image("mask_leaf", img_path, name, format=".png")
                diff_trunk_img = load_image("diff-trunk", img_path, name)
        else:
            self.report({"WARNING"}, "Diffuse images already created")

    # M A S K
    if at.mask_render:
        if not mask_exists or at.overwrite:
            print("")
            print("RENDERING MASKS:")
            mask_leaf_exists = os.path.isfile(img_path + "mask_leaf.png")
            mask_trunk_exists = os.path.isfile(img_path + "mask_trunk.png")
            mask_shadow_exists = os.path.isfile(img_path + "mask_shadow.png")
            render = bpy.context.scene.render

            # leaves
            if not mask_leaf_exists:
                print("Leaf mask...")
                print("")
                render.engine = "BLENDER_EEVEE_NEXT"

                # holdout bark
                shader_nodes = []
                for mat in bark_mats:
                    mat_output_node = unlink_output_node(compositor=False, material=mat)
                    holdout_node = mat.node_tree.nodes.new("ShaderNodeHoldout")
                    mat.node_tree.links.new(holdout_node.outputs[0], mat_output_node.inputs[0])
                    shader_nodes.append([holdout_node, mat_output_node])
                bpy.ops.wm.redraw_timer(type='DRAW_WIN_SWAP', iterations=1)

                alpha_link = node_tree.links.new(r_layers.outputs["Alpha"], output_node.inputs[4])

                bpy.ops.render.render()

                for i, mat in enumerate(bark_mats):
                    for node in shader_nodes[i]:
                        mat.node_tree.nodes.remove(node)

                node_tree.links.remove(alpha_link)
                rename_output(img_path + "mask_leaf", format=".png")

            # trunk mask
            if not mask_trunk_exists:
                print("Trunk mask...")
                print("")
                render.engine = "BLENDER_EEVEE_NEXT"
                leaf_object.hide_render = True
                alpha_link = node_tree.links.new(r_layers.outputs["Alpha"], output_node.inputs[5])

                bpy.ops.render.render()

                rename_output(img_path + "mask_trunk", format=".png")
                node_tree.links.remove(alpha_link)
                mask_trunk_exists = True

            # shadow
            if not mask_shadow_exists or at.overwrite:
                print("Shadow mask...")
                print("")
                render.engine = "CYCLES"
                collection = bpy.data.collections.new("leafcol")
                bpy.context.scene.collection.children.link(collection)
                collection.objects.link(leaf_object)
                layer_collection = bpy.context.view_layer.layer_collection.children[collection.name]
                layer_collection.indirect_only = True

                node_tree.links.new(r_layers.outputs["Image"], denoise_node.inputs[0])
                node_tree.links.new(r_layers.outputs["Denoising Normal"], denoise_node.inputs[1])
                node_tree.links.new(r_layers.outputs["Denoising Albedo"], denoise_node.inputs[2])
                node_tree.links.new(denoise_node.outputs[0], output_node.inputs[6])

                bpy.ops.render.render()

                bpy.data.collections.remove(collection)
                rename_output(img_path + "mask_shadow", format=".png")
            mask_shadow_img = load_image("mask_shadow", img_path, name, format=".png")

            if not mask_trunk_exists:
                rename_output(img_path + "mask_trunk", format=".png")
            mask_trunk_img = load_image("mask_trunk", img_path, name, format=".png")

            # load masks
            if not mask_leaf_img:
                mask_leaf_img = load_image("mask_leaf", img_path, name, format=".png")
            elif not diff_trunk_img:
                diff_trunk_img = load_image("diff-trunk", img_path, name)

            # combine masks
            print("Compositing final mask...")
            print("")
            # unlink all outputs
            for input in output_node.inputs:
                if input.links:
                    node_tree.links.remove(input.links[0])

            # make nodes
            diff_trunk_node = node_tree.nodes.new("CompositorNodeImage")
            mask_leaf_node = node_tree.nodes.new("CompositorNodeImage")
            mask_trunk_node = node_tree.nodes.new("CompositorNodeImage")
            mask_shadow_node = node_tree.nodes.new("CompositorNodeImage")
            bw_node = node_tree.nodes.new("CompositorNodeRGBToBW")
            multiply_node = node_tree.nodes.new("CompositorNodeMixRGB")
            combine_node = node_tree.nodes.new("CompositorNodeCombRGBA")

            # settings
            diff_trunk_node.image = diff_trunk_img
            mask_leaf_node.image = mask_leaf_img
            mask_trunk_node.image = mask_trunk_img
            mask_shadow_node.image = mask_shadow_img
            multiply_node.blend_type = "MULTIPLY"

            # links
            links = node_tree.links

            links.new(mask_shadow_node.outputs[0], bw_node.inputs[0])
            links.new(diff_trunk_node.outputs[0], multiply_node.inputs[1])
            links.new(bw_node.outputs[0], multiply_node.inputs[2])
            links.new(mask_trunk_node.outputs[0], combine_node.inputs[0])
            links.new(mask_leaf_node.outputs[0], combine_node.inputs[1])
            links.new(multiply_node.outputs[0], combine_node.inputs[2])
            links.new(combine_node.outputs[0], output_node.inputs[7])

            fast_render()

            # cleanup
            rename_output(img_path + "mask", format=".png")

            if at.remove_extra_masks:
                for image in ["mask_leaf", "mask_trunk", "mask_shadow"]:
                    os.remove(img_path + image + ".png")

            for node in [
                diff_trunk_node,
                mask_leaf_node,
                mask_trunk_node,
                mask_shadow_node,
                bw_node,
                multiply_node,
                combine_node,
            ]:
                comp_nodes.append(node)

            for input in output_node.inputs:
                if input.links:
                    links.remove(input.links[0])

        else:
            self.report({"WARNING"}, "Mask image already created")

    # rejoin objects
    bpy.ops.object.select_all(action='DESELECT')
    trunk_object.select_set(True)
    leaf_object.select_set(True)
    bpy.ops.object.join()

    # P R E V I E W
    if at.preview_render:

        print("")
        print("RENDERING PREVIEW")

        text_obj.data.body = "Rendering Preview..."
        bpy.ops.wm.redraw_timer(type='DRAW_WIN_SWAP', iterations=1)

        # import world
        for world in bpy.data.worlds:
            if world.name == AT_PREV_WORLD_NAME:
                break
        else:
            blend_path = str(helpers.get_resources_path(sub=AT_BLEND_NAME))
            with bpy.data.libraries.load(blend_path) as (data_from, data_to):
                for world in data_from.worlds:
                    if world == AT_PREV_WORLD_NAME:
                        data_to.worlds.append(world)
                        break

        context.scene.world = bpy.data.worlds[AT_PREV_WORLD_NAME]

        # load person image
        for img in bpy.data.images:
            if img.name == "person outline.png":
                person_img = img
                break
        else:
            person_img_path = str(helpers.get_resources_path(sub="images") / "person outline.png")
            person_img = bpy.data.images.load(person_img_path)

        # settings
        render = context.scene.render
        render.engine = "CYCLES"
        render.resolution_x = 256
        render.resolution_y = 256

        # Nodes

        links = node_tree.links

        # denoise node
        context.view_layer.cycles.denoising_store_passes = True
        links.new(r_layers.outputs["Image"], denoise_node.inputs[0])
        links.new(r_layers.outputs["Denoising Normal"], denoise_node.inputs[1])
        links.new(r_layers.outputs["Denoising Albedo"], denoise_node.inputs[2])

        # mix denoise
        mix_denoise_node = node_tree.nodes.new("CompositorNodeMixRGB")
        # noise
        samples = context.scene.cycles.samples
        mix_denoise_node.inputs[0].default_value = 0.8 if samples < 10 else 0.5
        comp_nodes.append(mix_denoise_node)

        links.new(r_layers.outputs["Image"], mix_denoise_node.inputs[1])
        links.new(denoise_node.outputs[0], mix_denoise_node.inputs[2])

        # person image
        person_img_node = node_tree.nodes.new("CompositorNodeImage")
        person_img_node.image = person_img
        comp_nodes.append(person_img_node)

        # scale
        scale_node = node_tree.nodes.new("CompositorNodeScale")
        TREE_HEIGHT = 10 / context.object.scale[0]
        PERSON_HEIGHT = 1.8  # in scene units aka meters
        scale_val = 1 / (TREE_HEIGHT/PERSON_HEIGHT)  # take the inverse
        scale_node.inputs[1].default_value = scale_val
        scale_node.inputs[2].default_value = scale_val
        comp_nodes.append(scale_node)

        links.new(person_img_node.outputs[0], scale_node.inputs[0])

        # translate
        translate_node = node_tree.nodes.new("CompositorNodeTranslate")
        # formula to put person in corner at 75px*256px resolution
        PERSON_RES = (75, 256)
        TREE_OFFSET = (256/2) * -1  # moves middle of person to the corner
        person_offset_x = (PERSON_RES[0] / 2) * scale_val  # moves person so that the corners align
        person_offset_y = (PERSON_RES[1] / 2) * scale_val
        translate_x = TREE_OFFSET + person_offset_x + 20  # Add arbitrary offset from side
        translate_y = TREE_OFFSET + person_offset_y
        translate_node.inputs[1].default_value = translate_x
        translate_node.inputs[2].default_value = translate_y
        comp_nodes.append(translate_node)

        links.new(scale_node.outputs[0], translate_node.inputs[0])

        # return aspect ratio
        return_aspect_node = node_tree.nodes.new("CompositorNodeMixRGB")
        return_aspect_node.inputs[0].default_value = 1
        comp_nodes.append(return_aspect_node)

        links.new(mix_denoise_node.outputs[0], return_aspect_node.inputs[1])
        links.new(translate_node.outputs[0], return_aspect_node.inputs[2])

        # get alpha map
        sep_rgba_node = node_tree.nodes.new("CompositorNodeSepRGBA")
        comp_nodes.append(sep_rgba_node)

        links.new(translate_node.outputs[0], sep_rgba_node.inputs[0])

        # return correct alpha
        set_alpha_node = node_tree.nodes.new("CompositorNodeSetAlpha")
        set_alpha_node.mode = "REPLACE_ALPHA"
        comp_nodes.append(set_alpha_node)

        links.new(return_aspect_node.outputs[0], set_alpha_node.inputs[0])
        links.new(sep_rgba_node.outputs[3], set_alpha_node.inputs[1])

        # add on top
        alpha_over_node = node_tree.nodes.new("CompositorNodeAlphaOver")
        comp_nodes.append(alpha_over_node)

        links.new(set_alpha_node.outputs[0], alpha_over_node.inputs[1])
        links.new(mix_denoise_node.outputs[0], alpha_over_node.inputs[2])
        links.new(alpha_over_node.outputs[0], output_node.inputs[8])

        # returnx
        bpy.context.scene.render.use_border = False
        bpy.ops.render.render()
        bpy.context.scene.render.use_border = True

        rename_output(img_path + "preview", format=".png")

    # remove nodes
    for node in comp_nodes:
        node_tree.nodes.remove(node)

    # relink composite node
    if output_link:
        node_tree.links.new(output_link, composite_link)

    try:
        reset_settings(settings)
    except TypeError:
        pass

    bpy.ops.object.delete()
    bpy.data.curves.remove(text_obj.data)
    bpy.data.scenes.remove(bpy.data.scenes["alpha_trees_scene"])
    context.window.scene = scene

    generate_lods(context, name.split("_")[0])

    tags = []
    for m in object.material_slots:
        m = m.material
        if m:
            if m.at.type == "FLOWER":
                tags.append("flowering")
            if m.at.type == "FRUIT":
                tags.append("fruiting")

    scale = 10/object.scale[0]
    helpers.write_metadata(context, name, scale, tags=tags)
    bpy.ops.alpha_tree.reload_previews()


# Utils
# ------------------------------------------------------------------------------------


# useful for seeing through the mess of errors
def xprint(message, len=10):
    if type(message) == list:
        for item in message:
            for _ in range(len):
                print(item)
    else:
        for _ in range(len):
            print(message)
