import imghdr
import json
import os
import struct
from os.path import join as osjoin
from pathlib import Path

import bpy
from bpy.app.handlers import persistent

from ..constants import (
    AT_BLEND_NAME,
    INPUT_NAMES,
    MOD_GROUP_VERSION,
    NODE_NAMES,
    TREE_METADATA_VERSION,
)
from ..weight_layers_addon.WL_functions import get_node_by_type, get_nodes_with_group
from ..weight_layers_addon.WL_functions import get_wl_obj as get_at_obj
from ..weight_layers_addon.WL_functions import remove_handler
from .. import __package__ as base_package
# UI


def centered_row(layout, use_opposite=False, fac=0.95):
    """Returns a row that is actually centered"""
    row = layout.row()
    # This split is here to center the labels,
    # cos for some reason they are very slightly to the left otherwise -_-
    split2 = row.split(factor=fac)
    row = split2.row()
    row2 = split2.row()
    row = row2 if use_opposite else row
    row.alignment = "CENTER"
    return row


def draw_tree_search(layout, data):
    col = layout.column(align=True)
    row = col.row(align=True)
    row.prop(
        data,
        "show_tree_search_settings",
        text="",
        icon="TRIA_DOWN" if data.show_tree_search_settings else "TRIA_RIGHT",
    )
    row.prop(data, "tree_search", text="", icon="VIEWZOOM")
    if data.show_tree_search_settings:
        box = col.box()
        col = box.column(align=True)
        col.prop(data, "tree_type")
        col.prop(data, "tree_subtype")
        col.prop(data, "tree_tag")
    return col


# Prefs


def get_resources_path(sub: str = "") -> Path:
    """Returns the path to the resources folder with the given subdirectory
    Args:
        sub (str, optional): A subdirectory in the resources folder. Defaults to "".
    """
    path = Path(__file__).resolve()
    path = path.parents[1].absolute()
    path = path / "resources" / sub
    return path


def get_prefs() -> bpy.types.AddonPreferences:
    """Returns a reference to the addon preferences

    Returns:
        bpy.types.AddonPreferences: The preferences of this addon
    """
    return bpy.context.preferences.addons[base_package].preferences


# Generator


def generate_tags(tree_name, scale):
    tags = []
    scale = float(scale)
    if scale > 15:
        if scale > 30:
            tags.append("massive")
        tags.append("big")
    elif scale < 5:
        tags.append("small")
    else:
        tags.append("medium")
    return tags


def write_metadata(context, tree_name, scale=0, type="", subtype="", tags: list = None):
    prefs = get_prefs()
    name_parts = tree_name.split("_")
    name = name_parts[0]
    path = bpy.path.abspath(osjoin(prefs.lib_path, "rendered_maps", name + "_data.json"))
    scale = float(scale)
    tags = tags if tags else []
    tags = list(set(tags))
    if not any(t in ["small", "medium", "big", "massive"] for t in tags):
        tags.extend(generate_tags(tree_name, scale))

    # set default tree attributes
    if not scale:
        # check that the name has a scale in it
        if len(name_parts) >= 2 and is_float(name_parts[1]):
            scale = float(name_parts[1]) * 10
        else:
            scale = 10
    if not type:
        if any(
            [type in name.lower() for type in ["pine", "spruce", "fir", "larch", "pinus", "abies", "picea", "larix"]]
        ):
            type = "conniferous"
        else:
            type = "shrub" if "small" in tags else "deciduous"
    if not subtype:
        for separator in [" ", "-", "."]:
            subtype = " ".join(tree_name.split(separator)[:-1])
            if subtype:
                break
        else:
            subtype = name

    data = {}
    data["version"] = TREE_METADATA_VERSION
    data["name"] = name
    data["scale"] = scale
    data["type"] = type
    data["subtype"] = subtype.lower()
    data["tags"] = tags

    with open(path, "w") as f:
        f.truncate(0)
        json.dump(data, f, indent=4)

    # convert old trees to use the new naming format (without the scale in the title, and with different suffixes)
    if len(name_parts) > 1:
        for subdir in ["", "med", "low"]:
            subpath = osjoin(prefs.lib_path, "rendered_maps", subdir)
            if not os.path.isdir(subpath):
                continue  # check whether LOD folders have been created

            for img in os.listdir(subpath):
                if name + "_" in img:
                    elements = img.split("_")
                    new_name = img

                    # remove scale
                    if is_float(name_parts[1]):
                        if not is_float(elements[1]):
                            continue
                        del elements[1]
                        new_name = "_".join(elements)

                    # update suffixes
                    changes = [
                        ("diff_leaf", "diff-leaf"),
                        ("diff_trunk", "diff-trunk"),
                        ("nor_leaf", "nor-leaf"),
                        ("nor_trunk", "nor-trunk"),
                    ]
                    for change in changes:
                        new_name = new_name.replace(change[0], change[1])

                    # update file name
                    if not os.path.exists(osjoin(subpath, new_name)):
                        os.rename(osjoin(subpath, img), osjoin(subpath, new_name))


def get_metadata(tree_name: str = "", reload: bool = False) -> dict:
    """Return a dictionary of the metadata for this tree.
    This info is cached when the function is first called, so after making a change,
    make sure to call it with reload = True.

    Args:
        tree_name (str): The name of the tree to get the metadata for
        reload (bool): Whether to reload the cache of all trees' metadata
    Returns:
        dict: dictionary containing the metadata for: [name, scale, type, tags],
        "" if no tree is provided.
    """
    prefs = get_prefs()

    if get_metadata._path != prefs.lib_path:
        reload = True

    tree_name = tree_name.split("_")[0]
    metadata_cache = {"": None} if reload else get_metadata._tree_metadata_cache
    if len(metadata_cache) > 1:
        return metadata_cache[tree_name] if tree_name else ""

    # load all metadata into cache.
    # This is needed because it can get a bit slow when there are a lot of trees in a library
    img_dir = Path(prefs.lib_path) / "rendered_maps"
    for prev in list(img_dir.glob("*preview.png")):
        # create metadata if it doesn't exist
        name_parts = prev.name.split("_")
        name = name_parts[0]
        json_path = img_dir / (name + "_data.json")
        if not json_path.exists:
            if len(name_parts) >= 3 and is_float(name_parts[1]):
                scale = float(name_parts[1])
            else:
                scale = 10

            write_metadata(
                bpy.context,
                name,
                scale,
            )

        with open(json_path, "r") as f:
            metadata = json.load(f)

        metadata_cache[metadata["name"]] = metadata

    get_metadata._tree_metadata_cache = metadata_cache
    get_metadata._path = prefs.lib_path
    return metadata_cache[tree_name] if tree_name else ""


get_metadata._tree_metadata_cache = {"": None}
get_metadata._path = ""


def get_all_metadata(reload=False):
    """Returns a dictionary containing the metadata for all trees in the current library
    Args:
        reload (bool, optional): Whether to reload the cached metadata. Defaults to False.
    """
    prefs = get_prefs()

    if get_metadata._path != prefs.lib_path or len(get_metadata._tree_metadata_cache) < 1 or reload:
        get_metadata(tree_name="", reload=True)
    return get_metadata._tree_metadata_cache


# Importer


def get_other_maps(context, preview_name: str, quality: str) -> list[bpy.types.Image]:
    """Gets all of the image maps for a given preview image name
    Args:
        preview_name (str): The name of the tree preview image
        quality (str): The quality of the image maps to get (out of "high", "med", "low")
    Returns:
        list[bpy.types.Image]: a list of Blender image data blocks corresponding all of the tree maps
    """
    images_path = osjoin(get_prefs().lib_path, "rendered_maps")
    pack_images = get_prefs().auto_pack_images

    if quality != "high":
        images_path = osjoin(images_path, quality)

    images = []

    for suffix in ["diff-leaf.jpg", "diff-trunk.jpg", "nor-leaf.jpg", "nor-trunk.jpg", "mask.png"]:
        new_name = preview_name.replace("preview.png", suffix)
        for img in bpy.data.images:
            if new_name in img.name:
                if img.at_quality == quality:
                    break
        else:
            img = bpy.data.images.load(osjoin(images_path, new_name))
            img.name = new_name + f"_{quality}"
            img.at_quality = quality
            if pack_images:
                img.pack()

        images.append(img)

    return images


def add_constraints(self, obj: bpy.types.Object, on):
    # convert = {
    #     "POS_X": "TRACK_X",
    #     "POS_Y": "TRACK_Y",
    #     "POS_Z": "TRACK_Z",
    #     "NEG_X": "TRACK_NEGATIVE_X",
    #     "NEG_Y": "TRACK_NEGATIVE_Y",
    #     "NEG_Z": "TRACK_NEGATIVE_Z",
    # }

    if obj.constraints:
        track_const = obj.constraints.get("AT_TRACK_CONSTRAINT")
        limit_const = obj.constraints.get("AT_LIMIT_CONSTRAINT")

        if on:
            if track_const:
                track_const.target = self.camera
                track_const.mute = False
            if limit_const:
                limit_const.mute = False
        else:
            if track_const:
                obj.constraints.remove(track_const)
            if limit_const:
                obj.constraints.remove(limit_const)
    else:
        if on:
            track_const = obj.constraints.new("TRACK_TO")
            track_const.name = "AT_TRACK_CONSTRAINT"
            track_const.track_axis = "TRACK_NEGATIVE_Z" if self.invert_track else "TRACK_Z"
            track_const.target = self.camera
            track_const.owner_space = "LOCAL"

            limit_const = obj.constraints.new("LIMIT_ROTATION")
            limit_const.name = "AT_LIMIT_CONSTRAINT"
            limit_const.use_limit_x = obj.at.is_at
            limit_const.use_limit_y = obj.at.is_at


# Biomes


def get_biome_vars(self, context) -> list:
    """Returns [biome_list, list_settings, b_index, slot_node, biome]"""
    obj = get_at_obj(context)
    biome_list = obj.at.biome_list
    list_settings = obj.at
    b_index = list_settings.b_index
    if obj:
        if obj.at.biome_list:
            biome = biome_list[b_index]
        else:
            biome = None

    mod_group = get_mod_group(obj)
    slot_node = mod_group.nodes.get(str(b_index)) if mod_group else None

    return biome_list, list_settings, b_index, slot_node, biome


def get_default_obj(self, context, control_node: bpy.types.GeometryNode) -> bpy.types.Object:
    """Returns the default object that is scattered when no tree is selected for a particle slot
    Args:
        control_node (bpy.types.GeometryNode): The control node for the biome that the default obj wil be added to
        (used to get the collection)

    Returns:
        bpy.types.Object: The default object in the biomes collection
    """
    plane_index = 0
    curr_coll = context.collection
    orig_obj = context.active_object

    collection = get_input(control_node, "collection").default_value

    # get index
    while True:
        try:
            _ = collection.objects["AT_default_obj_" + str(plane_index)]
            plane_index += 1
        except KeyError:
            break

    for obj in bpy.data.objects:
        if obj.name == "AT_default_obj_" + str(plane_index):
            default_plane = obj
            break
    else:
        bpy.ops.mesh.primitive_plane_add()
        default_plane = bpy.context.active_object
        default_plane.name = "AT_default_obj_" + str(plane_index)

        verts = default_plane.data.vertices
        size = 0.5
        verts[0].co = (-size, 0, 0)
        verts[1].co = (size, 0, 0)
        verts[2].co = (-size, 0, 10)
        verts[3].co = (size, 0, 10)

        default_plane.track_axis = "POS_Z"
        default_plane.location = (0, 0, 0)
        curr_coll.objects.unlink(default_plane)

    context.view_layer.objects.active = orig_obj

    return default_plane


def get_at_node_group(name: str, type: str = "BIOME") -> bpy.types.GeometryNodeGroup:
    """Returns a new version of either the biome settings or tree settings node group.
    If a version is not present in the current blend file, it will be appended from base_scene.blend,
    otherwise, the version that is already there will be copied."""

    path = Path(__file__).resolve()
    path = path.parents[1].absolute()
    path = path / "resources" / AT_BLEND_NAME
    # path = imp_functions.get_resources_path(sub=AT_BLEND_NAME)
    settings_groups = [g for g in bpy.data.node_groups if name in g.name]
    if settings_groups:
        if type == "BIOME":
            settings_group = settings_groups[0].copy()
            # reset at settings
            settings_group.at.tree_list.clear()
            settings_group.at.t_index = 0
        else:
            settings_group = settings_groups[0]
    else:
        # append group
        with bpy.data.libraries.load(str(path)) as (data_from, data_to):
            for ng in data_from.node_groups:
                if name in ng:
                    break
            data_to.node_groups.append(ng)
        settings_group = data_to.node_groups[0]
    return settings_group


def get_all_bsettings():
    """Returns all biome settings as node groups in the blend file"""
    return [ng for ng in bpy.data.node_groups if NODE_NAMES["slot"] in ng.name]


# Trees


def get_tree_vars(self, context) -> list:
    """Returns [tree_list, settings_group, t_index, tree]"""
    settings_group = get_biome_vars(self, context)[-2].node_tree
    t_index = settings_group.at.t_index
    tree_list = settings_group.at.tree_list

    if settings_group:
        if settings_group.at.tree_list:
            tree = tree_list[t_index]
        else:
            tree = None

    return tree_list, settings_group, t_index, tree


# Vgroups


def get_biome_slot_mod_keys(biome_settings: bpy.types.GeometryNodeGroup, modifier: bpy.types.Modifier) -> list[str]:
    """Returns the value input keys for the alpha trees modifier, from which the value can be accessed with modifier[key]
    To get other attributes, just add "_attribute_name" or "_use_attribute" to the end.

    Args:
        stack_group (bpy.types.GeometryNodeGroup): The node group of the layer stack that you want the keys for
        modifier (bpy.types.Modifier): The modifier to get the keys for
    Returns:
        list[str]: A list of keys, depends on the order of the inputs on the stack node, usually [density, scale]
    """
    ng = modifier.node_group
    settings_node = get_nodes_with_group(ng.nodes, biome_settings, list=True)[0]
    group_input = get_node_by_type(ng, "GROUP_INPUT")
    filtered = [k for k in modifier.keys() if "use_attribute" not in k and "attribute_name" not in k]
    names = []
    for socket in settings_node.inputs[1:]:  # exclude Geometry input
        if socket.links and socket.links[0].from_node == group_input:
            in_socket = socket.links[0].from_socket
            idx = list(group_input.outputs).index(in_socket) - 1
            names.append(filtered[idx])
    return names


# Misc
def get_bl_list_index(item: bpy.types.PropertyGroup) -> int:
    """Get the index of any collection property item

    Args:
        item (bpy.types.PropertyGroup): The collection property item
    Returns:
        int: The index of the property in the collection
    """
    slot_path = "[".join(item.path_from_id().split("[")[:-1])
    list_prop = item.id_data.path_resolve(slot_path)
    return list(list_prop).index(item)


def get_image_size(fname: str) -> list[int, int]:
    """Determine the image type of fhandle and return its size.
    from draco"""

    # from this answer: https://stackoverflow.com/a/20380514,
    # because frankly, I don't have the braincells to try to understand this...
    # I need to use it because PIL isn't in the default library (for some reason)
    # and I don't really want the hassle of getting users to have to install external libraries,
    # even if it is only one button.

    with open(fname, "rb") as fhandle:
        head = fhandle.read(24)
        if len(head) != 24:
            return
        elif imghdr.what(fname) == "jpeg":
            try:
                fhandle.seek(0)  # Read 0xff next
                size = 2
                ftype = 0
                while not 0xC0 <= ftype <= 0xCF:
                    fhandle.seek(size, 1)
                    byte = fhandle.read(1)
                    while ord(byte) == 0xFF:
                        byte = fhandle.read(1)
                    ftype = ord(byte)
                    size = struct.unpack(">H", fhandle.read(2))[0] - 2
                # We are at a SOFn block
                fhandle.seek(1, 1)  # Skip `precision' byte.
                height, width = struct.unpack(">HH", fhandle.read(4))
            except Exception:  # IGNORE:W0703
                return
        else:
            return
        return width, height


def get_mat_control_node(self, context, tree) -> bpy.types.Node:
    if isinstance(tree, bpy.types.Object):
        object = tree
    else:
        object = tree.tree_obj

    try:
        control_node = get_node(object.material_slots[0].material.node_tree, "mat_control")
    except (AttributeError, IndexError, KeyError):
        control_node = None
    return control_node


def get_mod_group(object: bpy.types.Object, modifier: bool = False) -> bpy.types.GeometryNodeGroup:
    """returns the Alpha Trees geometry nodes modifier on this object,
    or None if not present. if modifier is False, returns the node group
    of the modifier."""
    for m in object.modifiers:
        if m.type == "NODES" and "AT_GEO_NODE" in m.name:
            node_modifier = m
            break
    else:
        node_modifier = None

    if node_modifier:
        mod_group = node_modifier.node_group
        # if not mod_group:
        if mod_group.users > 1:
            # check to see if being called from ui (can't modify blend data while drawing)
            update_name = True
            try:
                mod_group.name = mod_group.name
            except AttributeError:
                update_name = False
            if update_name:
                mod_group = node_modifier.node_group.copy()
                mod_group.name = "AT_BIOME_LIST_" + get_at_obj(bpy.context).name
                mod_group.at.version = MOD_GROUP_VERSION
                node_modifier.node_group = mod_group

    return node_modifier if modifier or not node_modifier else node_modifier.node_group


def get_list_index(item, type="int"):
    """Returns the index of an item in a blender CollectionProperty
    as either an int or a string. Doesn't work on all collection properties,
    not sure why yet"""
    index = item.path_from_id()[-2]  # should give the string of the item's index
    return int(index) if type == "int" else index


def get_node(tree, type):
    """Returns the correct node according to the type as defined in constants.py"""
    if not isinstance(tree, bpy.types.GeometryNodeGroup):
        nodes = tree.nodes
    else:
        nodes = tree.node_tree.nodes
    return nodes[NODE_NAMES[type]]


def get_input(node, type):
    """Returns the correct input according to the type as defined in constants.py"""
    try:
        return node.inputs[INPUT_NAMES[type]]
    except KeyError:
        return None


def get_preview_maps() -> list[str]:
    dir = osjoin(get_prefs().lib_path, "rendered_maps", "")
    preview_maps = []

    for file in os.listdir(dir):
        if os.path.isfile(osjoin(dir, file)):
            if "preview" in str(file).split("_")[-1]:
                preview_maps.append(file)

    return preview_maps


def is_float(string):
    try:
        float(string)
        return True
    except ValueError:
        return False


@persistent
def at_change_cam_handler(scene):
    for obj in bpy.data.objects:
        mod_group = get_mod_group(obj)
        if mod_group:
            for node in mod_group.nodes:
                if node.type == "GROUP" and node.node_tree.at.type == "SETTINGS":
                    # update node sockets
                    control_node = get_node(node.node_tree, "control")
                    cam_socket = get_input(control_node, "camera")
                    if node.node_tree.at.use_scene_cam:
                        if cam_socket and cam_socket.default_value != scene.camera:
                            cam_socket.default_value = scene.camera
                    socket = get_input(control_node, "cam_fov")
                    if cam_socket.default_value and socket.default_value != cam_socket.default_value.data.angle:
                        socket.default_value = cam_socket.default_value.data.angle
                    socket = get_input(control_node, "cam_res_x")
                    if socket.default_value != scene.render.resolution_x:
                        socket.default_value = scene.render.resolution_x
                    socket = get_input(control_node, "cam_res_y")
                    if socket.default_value != scene.render.resolution_y:
                        socket.default_value = scene.render.resolution_y

                elif node.type == "GROUP" and NODE_NAMES["single_tree"] in node.name:
                    if obj.at.use_scene_cam:
                        socket = get_input(node, "single_tree_cam")
                        if socket and socket.default_value != scene.camera:
                            socket.default_value = scene.camera


def register():
    # For some reason, the depsgraph isn't always updated when the frame is changed???
    # I guess it makes sense...
    bpy.app.handlers.frame_change_pre.append(at_change_cam_handler)
    bpy.app.handlers.depsgraph_update_pre.append(at_change_cam_handler)


def unregister():
    remove_handler(at_change_cam_handler, type="frame_change_pre")
    remove_handler(at_change_cam_handler, type="depsgraph_update_pre")
