import bpy
import bpy.utils.previews
import os
from os.path import join as osjoin
from mathutils import Vector

from .helper_functions import get_resources_path, get_image_size, at_change_cam_handler,\
    get_other_maps, get_node, get_input, get_metadata, get_prefs, add_constraints, get_mat_control_node
from ..info import draw_info
from ..constants import AT_MAT_NAME, AT_BLEND_NAME, NODE_NAMES, MAT_VERSION, TREE_OBJ_VERSION


def get_icon(name, type="icon", size="large", data=None):
    from ..icons import get_icon as icons_get_icon
    return icons_get_icon(name, type, size, data)


# UI
# --------------------------------------------------------------------------


def draw_single_material_section(context: bpy.types.Context, layout: bpy.types.UILayout, obj: bpy.types.Object):
    at = context.scene.alpha_trees
    prefs = get_prefs()
    control_node = get_mat_control_node(None, context, obj)
    if control_node:
        layout.separator(factor=0)
        col = layout.column(align=True)
        boxcol = col.box().column(align=True)
        row = boxcol.row(align=True)

        draw_open_panel_button(
            context,
            row,
            at,
            at.show_material_settings,
            "show_material_settings",
            "MATERIAL",
        )
        row.label(text="Material")

        row2 = row.row()
        draw_info(row2, "Material settings", "material_settings")

        row.operator(
            "alpha_tree.copy_settings_to_single",
            text="",
            icon="COPYDOWN",
            emboss=False,
        ).type = "MATERIAL"

        if at.show_material_settings:
            box = col.box()
            row = box.row(align=True)
            row.template_ID(
                obj.at,
                "material",
                live_icon=prefs.use_mat_icons,
                unlink="alpha_tree.remove_at_material",
            )

            diff_name = get_node(obj.active_material.node_tree, "leaf_col").image.name
            # remove suffixes
            diff_name = diff_name[:diff_name.index(".jpg") + 4]
            if os.path.isfile(os.path.join(prefs.lib_path, "rendered_maps", "low", diff_name)):
                row = box.row(align=True)
                row.prop(obj.active_material.at, "quality", text="")
            else:
                row = box.row(align=True)
                prev_name = diff_name.replace("diff-leaf.jpg", "preview.png")
                row.operator("alpha_tree.generate_lods").tree_name = prev_name
                draw_info(row, "Create LODs", "create_lods", emboss=True)

            boxcol = col.column(align=True)
            draw_material_settings(context, boxcol, control_node)

    else:
        if obj:
            if obj.material_slots and obj.at.is_at:
                if not obj.active_material:
                    layout.separator()
                    col = layout.column(align=True)
                    boxcol2 = col.box().column(align=True)
                    row = boxcol2.row(align=True)
                    row.template_ID(
                        obj.at,
                        "material",
                        live_icon=prefs.use_mat_icons,
                        new="alpha_tree.new_at_material",
                    )


def draw_material_settings(context, layout, node):
    sockets = [socket for socket in node.inputs if socket.enabled and not socket.is_linked]
    category = ""
    for socket in sockets:
        cat = socket.name.split("_")[0]
        name = socket.name.split("_")[1].capitalize()
        if cat != category:
            category = cat
            box = layout.box()
            row = box.row(align=True)
            row.alignment = "CENTER"
            row.label(text=cat.capitalize())
            col = box.column(align=True)

        socket.draw(context, col, node, name)


def draw_preview_enum(self, context, layout, operators: list, enum_data, enum_name):

    row = layout.row(align=True)
    col = row.column(align=True)
    cc = col.column(align=True)
    scale_y = 6

    cc.scale_y = 1
    cc.operator("alpha_tree.select_random_tree", text="", icon_value=get_icon("random.png"))

    cc = col.column(align=True)
    cc.scale_y = scale_y - 1
    cc.scale_x = 0.92
    cc.operator(operators[0].bl_idname, text="", icon="TRIA_LEFT").forward = False
    if len(operators) > 1:
        cc = col.column(align=True)
        cc.scale_y = 1
        cc.operator(operators[1].bl_idname, text="", icon="FILE_REFRESH")

    row.template_icon_view(enum_data, enum_name, scale=scale_y + 1, scale_popup=7, show_labels=True)

    col = row.column(align=True)

    cc = col.column(align=True)
    cc.scale_y = 1
    cc.operator("alpha_tree.edit_tree_settings", text="", icon="GREASEPENCIL")

    cc = col.column(align=True)
    cc.scale_y = scale_y - 1
    cc.scale_x = 0.92
    cc.operator(operators[0].bl_idname, text="", icon="TRIA_RIGHT").forward = True
    if len(operators) > 2:
        cc = col.column(align=True)
        cc.scale_y = 1
        cc.operator(operators[2].bl_idname, text="", icon_value=get_icon("multi_import.png"))


def draw_text(layout, lines):
    for line in lines:
        layout.label(text=line)


def get_tree_icon(context, tree):
    if get_prefs().use_mat_icons:
        icon_val = tree.tree_obj.active_material.preview.icon_id
    else:
        icon_val = get_icon(tree.selected_tree, type="preview")
    return icon_val


def draw_open_panel_button(context, layout, data, bool, bool_str, icon):
    if get_prefs().use_arrows:
        layout.prop(data, bool_str, text="", icon="TRIA_DOWN" if bool else "TRIA_RIGHT")
    else:
        if isinstance(icon, int):
            layout.prop(data, bool_str, text="", icon_value=icon)
        else:
            layout.prop(data, bool_str, text="", icon=icon)


# Callbacks
# --------------------------------------------------------------------------


def material_poll(self, object):
    if object.at.is_at:
        return True
    return False


def material_update(self, context):
    obj = self.id_data
    obj.material_slots[0].material = self.material

    # scale to correct ratio
    if self.material:
        diff_leaf_image = get_node(self.material.node_tree, "leaf_col").image
        verts = obj.data.vertices
        if len(verts) == 4 and diff_leaf_image:
            ratio = diff_leaf_image.size[1] / diff_leaf_image.size[0]
            width = (Vector(verts[0].co) - Vector(verts[1].co)).length

            verts[2].co = (verts[0].co[0], verts[0].co[1], width * ratio)
            verts[3].co = (verts[1].co[0], verts[1].co[1], width * ratio)


def camera_poll(self, object):
    for obj in bpy.context.scene.objects:
        if obj == object and obj != bpy.context.object:
            return True
    return False


def camera_update(self, context):
    obj = self.id_data

    # for single object or whole psystem
    if not isinstance(obj, bpy.types.Object):
        psetting = obj
        coll = psetting.instance_collection
        if not coll:
            return
        for obj in coll.objects:
            add_constraints(self, obj, self.camera)
    else:
        add_constraints(self, obj, self.camera)


def invert_track_update(self, context):
    obj = self.id_data

    if not isinstance(obj, bpy.types.Object):
        psetting = obj
        coll = psetting.instance_collection
        if not coll:
            return
        for obj in coll.objects:
            obj.at.invert_track = self.invert_track
    else:
        if obj.constraints:
            track_const = obj.constraints.get("AT_TRACK_CONSTRAINT")

            if track_const:
                track_const.track_axis = "TRACK_NEGATIVE_Z" if self.invert_track else "TRACK_Z"


def rotation_offset_update(self, context):
    obj = self.id_data

    # for single object or whole psystem
    if not isinstance(obj, bpy.types.Object):
        psetting = obj
        coll = psetting.instance_collection
        if not coll:
            return
        for obj in coll.objects:
            obj.at.rotation_offset = self.rotation_offset
    else:
        obj.rotation_euler[2] = self.rotation_offset


# Functions
# --------------------------------------------------------------------------


def get_tree_quality_items(self, context):
    items = []
    prefs = get_prefs()

    high_path = osjoin(bpy.path.abspath(prefs.lib_path), "rendered_maps")
    med_path = osjoin(high_path, "med")
    low_path = osjoin(high_path, "low")
    if isinstance(self.id_data, bpy.types.Material):
        try:
            mat = self.id_data
            img_name = get_node(mat.node_tree, "leaf_col").image.name
            tree_name = img_name[:img_name.index(".jpg") + 4]
        except AttributeError:
            img_name = ""
            tree_name = "default.png"

    else:
        tree_name = self.selected_tree

    if tree_name and "default.png" in tree_name or not os.path.isdir(med_path):
        items = [
            ("high", "high", "high"),
            ("med", "med", "med"),
            ("low", "low", "low"),
        ]
        return items

    for type, path in zip(["high", "med", "low"], [high_path, med_path, low_path]):
        for img_file in os.listdir(path):
            if tree_name.replace("preview.png", "diff-leaf.jpg") == img_file:
                break
        else:
            if prefs.show_dev_options:
                print("Couldn't find tree!")
            continue
        path = osjoin(path, img_file)

        size = max(get_image_size(path))

        name = "medium" if type == "med" else type
        name = name.capitalize() + f" quality ({str(size)}px)"
        items.append((type, name, name))

    return items


def import_material(blend_path):
    with bpy.data.libraries.load(blend_path) as (_, data_to):
        data_to.materials.append(AT_MAT_NAME)
    data_to.materials[0].at.version = MAT_VERSION
    return (data_to.materials[0])


def switch_images(images, nodes):
    prev_images = []
    labels = [
        NODE_NAMES["leaf_col"], NODE_NAMES["trunk_col"], NODE_NAMES["leaf_nor"], NODE_NAMES["trunk_nor"],
        NODE_NAMES["mask"]
    ]

    for node in nodes:
        if node.type == "TEX_IMAGE":
            for label, img in zip(labels, images):
                if node.label == label:
                    prev_images.append(node.image)
                    node.image = img
                    break

    for img in prev_images:
        if img:
            if img.users == 0:
                bpy.data.images.remove(img)


def import_tree(image_path, copy_mat, quality):

    preview_name = os.path.basename(image_path)
    tree_name = preview_name.split("_")[0]
    tree_size = get_metadata(preview_name)["scale"]

    # get duplicates
    obj_index = 0
    while True:
        try:
            _ = bpy.data.objects[tree_name + "_" + str(obj_index)]
            obj_index += 1
        except KeyError:
            break

    tree_name = tree_name + "_" + str(obj_index)

    images = get_other_maps(bpy.context, preview_name, quality)

    ratio = images[0].size[1] / images[0].size[0]
    height = 2 * ratio
    bpy.ops.mesh.primitive_plane_add()
    plane = bpy.context.active_object
    plane.name = tree_name

    verts = plane.data.vertices
    verts[0].co = (-1, 0, 0)
    verts[1].co = (1, 0, 0)
    verts[2].co = (-1, 0, height)
    verts[3].co = (1, 0, height)

    target_height = float(tree_size)
    scale = target_height / height

    plane.scale = (scale, scale, scale)
    plane.track_axis = "POS_Z"

    bpy.ops.object.material_slot_add()
    mat_slot = plane.material_slots[0]

    for material in bpy.data.materials:
        if material.name == AT_MAT_NAME:
            base_tree_material = material
            break
    else:
        blend_path = str(get_resources_path(sub=AT_BLEND_NAME))
        base_tree_material = import_material(blend_path)

    if copy_mat:
        for material in bpy.data.materials:
            if material.name == "AT_" + tree_name:
                tree_material = material
                break
        else:
            tree_material = base_tree_material.copy()
            tree_material.name = "AT_" + tree_name
    else:
        tree_material = base_tree_material.copy()
        tree_material.name = "AT_" + tree_name

    tree_material.preview_render_type = "FLAT"
    tree_material.preview.reload()
    mat_slot.material = tree_material
    tree_material.at.is_at = True
    plane.at.is_at = True
    plane.at.material = tree_material
    plane.at.version = TREE_OBJ_VERSION

    tree_material.at.quality = quality

    # Switch images

    nodes = tree_material.node_tree.nodes
    switch_images(images, nodes)

    return plane


def use_scene_cam_update(self, context):

    if isinstance(self.id_data, bpy.types.GeometryNodeTree):
        control_node = get_node(self.id_data, "control")
        socket = get_input(control_node, "camera")
        if self.use_scene_cam:
            self.prev_camera = socket.default_value
            at_change_cam_handler(context.scene)
        else:
            socket.default_value = self.prev_camera
    else:
        if self.use_scene_cam:
            self.prev_camera = self.camera
            at_change_cam_handler(context.scene)
        else:
            self.camera = self.prev_camera
