import bpy
from bpy.props import CollectionProperty, StringProperty, BoolProperty, EnumProperty, IntProperty

from pathlib import Path
import os

from .t3dn_bip import previews
from .t3dn_bip.ops import InstallPillow
from .functions.helper_functions import get_metadata, get_resources_path, get_prefs, get_all_metadata
from .weight_layers_addon.WL_functions import remove_handler
"""This uses the 3D Ninjas Blender Image Preview (BIP) Library to load previews quickly.
It's amazing, and you should absolutely check it out: https://github.com/3dninjas/3dn-bip
"""

# This is a hack to get stop the blender dynamic enum bug where names will not display correctly.
# I don't really understand the solution, but it took a lot of searching to find,
# massive thanks to @coreprocess on github for the solution:
# https://github.com/3dninjas/3dn-bip/issues/51#issuecomment-925001999


def make_item(prev_dict, id, name, descr, preview_id, uid):
    lookup = f"{str(id)}\0{str(name)}\0{str(descr)}\0{str(preview_id)}\0{str(uid)}"
    if lookup not in prev_dict:
        prev_dict[lookup] = (id, name, descr, preview_id, uid)
    return prev_dict[lookup]


tree_prev_dict = dict()
weight_prev_dict = dict()


def get_tree_previews(data=None, high_res_tree="", reload=False):
    """Gets all previews in the library directory, and returns them as blender enum items.
    Only the currently selected tree is loaded at full 256p x 256p, so it doesn't cause lag when
    selecting a tree (aka when all tree previews are being shown). The rest are loaded at 128p x 128p.

    To improve performance with large tree libraries, the preview items are cached and then refreshed when a property
    such as the tree search is changed.

    Args:
        high_res_tree (str, optional): The name of a tree to get at high resolution. Used for showing
            large single trees.
    """

    # check whether previews have been cached
    previews = []
    if not reload and data.prev_cache_small and data.prev_cache_large:
        mixed_previews = data.cache_to_enum_items(suffix="small")
        if high_res_tree and data:
            for i, prev in enumerate(mixed_previews):
                if prev[0] == high_res_tree:
                    item = mixed_previews[i]
                    item = (item[0], item[1], item[2], data.prev_cache_large[i].icon_id, item[-1])
                    mixed_previews[i] = item
        return mixed_previews

    if not pcolls.keys():
        return

    # add default icon
    default_path = os.path.join(get_resources_path(sub="images"), "default.png")
    icon = pcolls["AT_tree_previews_large"].load_safe("default.png", default_path, "IMAGE").icon_id
    previews.append(make_item(tree_prev_dict, "default.png", "", "default.png", icon, 0))

    for suffix, preview_list in zip(["small", "large"], [previews.copy(), previews.copy()]):
        # add all tree icons in the biome directory
        folder = Path(get_prefs().lib_path) / "rendered_maps"
        paths = list(folder.glob('*preview.png'))
        all_metadata = get_all_metadata()

        for index, path in enumerate(paths):

            name = path.name.replace("_preview.png", "")
            metadata = all_metadata[name]

            search = data.tree_search.lower() if data else ""
            if not search or search in path.name.lower()\
                or any([search in t for t in metadata["tags"]])\
                or search in metadata["type"].lower()\
                or search in metadata["subtype"].lower() or path.name == high_res_tree:

                type = data.tree_type if data else "ALL"
                subtype = data.tree_subtype if data else "ALL"
                tag = data.tree_tag if data else "ALL"
                if all([
                        type == metadata["type"] or type == "ALL",
                        subtype == metadata["subtype"] or subtype == "ALL",
                        tag in metadata["tags"] or tag == "ALL",
                ]) or path.name == high_res_tree:

                    icon = pcolls["AT_tree_previews_" + suffix].load_safe(path.name, str(path), 'IMAGE')

                    scale = round(metadata["scale"])
                    name = name + f" ({scale}m)"
                    preview_list.append(make_item(tree_prev_dict, path.name, name, path.name, icon.icon_id, index + 1))

        if not data:
            continue

        cache = data.prev_cache_small if suffix == "small" else data.prev_cache_large
        try:
            cache.clear()
        except AttributeError as e:
            print("Alpha Trees Error: " + str(e))
            continue
        for item in preview_list:
            cache_item = cache.add()
            cache_item.id = item[0]
            cache_item.name = item[1]
            cache_item.icon_id = item[3]
            cache_item.uid = item[-1]

    return previews


def get_tree_enum_previews(self, context: bpy.types.Context):
    """Return enum items with correct previews
    """
    # accessing self.selected_tree causes an infinite loop,
    # so use a property that is updated every time the tree is changed
    current_tree = ""
    try:
        current_tree = self["tree_copy"]
    except KeyError:  # property has not been updated/set yet
        pass
    prevs = get_tree_previews(
        data=self,
        high_res_tree=current_tree,
    )
    return prevs


def selected_tree_icon_update(self):
    self["tree_copy"] = self.selected_tree


def get_weight_previews(self, context: bpy.types.Context):
    preview_list = []
    folder = get_resources_path(sub="images")
    paths = folder.glob('*weight*')

    for index, path in enumerate(list(paths)):
        icon = pcolls["AT_weight_previews"].load_safe(path.name, str(path), 'IMAGE')
        identifier = path.name.removeprefix("weight_").removesuffix(".png")
        name = identifier.replace("_", " ").capitalize()
        preview_list.append(make_item(weight_prev_dict, identifier, name, name, icon.icon_id, index))
    return preview_list


def get_icon(name, type="icon", size="large", data=None) -> int:
    """Return the icon id of the given icon name and type
    Args:
        type (str, optional): Enum in ["icon", "preview", "weight"]. Defaults to "icon".
        size (str, optional): the resolution of the preview. Only applies to tree previews. Defaults to "large"
    Returns:
        int: The icon_id of the given icon
    """
    if type == "icon":
        pcoll = pcolls["AT_icons"]
    elif type == "preview":
        pcoll = pcolls["AT_tree_previews_" + size]
        try:
            _ = pcoll[name]
        except KeyError:
            if size == "large":
                get_tree_previews(data=data, high_res_tree=name)
            else:
                get_tree_previews(data=data)
    elif type == "weight":
        pcoll = pcolls["AT_weight_previews"]
    return pcoll[name].icon_id


def reload_all_previews(self, context):
    """Force reload all tree previews. Useful if library is updated or new trees are added or removed"""
    get_metadata(reload=True)
    # update tree preview cache for single tree importing.
    get_tree_previews(data=context.scene.alpha_trees, reload=True)
    # update tree preview caches for all existing trees in this file.
    try:
        for ng in bpy.data.node_groups:
            if ng.type == "GEOMETRY":
                for tree in ng.at.tree_list:
                    get_tree_previews(data=tree, reload=True)
    except AttributeError:
        # if tries to run when registering and restrict data is active.
        pass


# Settings
class PreviewCacheItems(bpy.types.PropertyGroup):
    """Used to store info about each preview when it is cached."""

    id: StringProperty()
    name: StringProperty()
    icon_id: IntProperty()
    uid: IntProperty()


# callbacks
def get_metadata_items(type):
    metadata = get_all_metadata()
    types = set()
    items = [("ALL", "All", "show all")]
    for m in metadata.values():
        if not m:
            continue
        if isinstance(m[type], list):
            for v in m[type]:
                types.add(v)
        else:
            types.add(m[type])
    if types:
        items.extend([(t, t.capitalize(), t) for t in types])
    return items


def get_tree_type_items(self, context):
    return get_metadata_items(type="type")


def get_tree_subtype_items(self, context):
    return get_metadata_items(type="subtype")


def get_tree_tags_items(self, context):
    return get_metadata_items(type="tags")


def tree_search_update(self, context):
    get_tree_previews(
        data=self,
        high_res_tree=self.selected_tree,
        reload=True,
    )


class PreviewSettingsParent(bpy.types.PropertyGroup):
    """Parent class that provides properties for filtering tree previews.
    Used for TreeListItem and ATObjectSettings"""

    # Tree search
    def cache_to_enum_items(self, suffix):
        list = getattr(self, "prev_cache_" + suffix)
        items = []
        for i, item in enumerate(list):
            items.append(make_item(tree_prev_dict, item.id, item.name, item.id, item.icon_id, item.uid))
        return items

    prev_cache_small: CollectionProperty(type=PreviewCacheItems)
    prev_cache_large: CollectionProperty(type=PreviewCacheItems)

    tree_search: StringProperty(
        name="Tree search",
        description="Only show trees with containing this string",
        update=tree_search_update,
    )

    show_tree_search_settings: BoolProperty(name="Show search settings",)

    tree_type: EnumProperty(
        items=get_tree_type_items,
        name="Type",
        update=tree_search_update,
    )

    tree_subtype: EnumProperty(
        items=get_tree_subtype_items,
        name="Subtype",
        update=tree_search_update,
    )

    tree_tag: EnumProperty(
        items=get_tree_tags_items,
        name="Tags",
        update=tree_search_update,
    )

    pass


class ALPHATREE_OT_install_pillow(InstallPillow):
    bl_idname = "alpha_tree.install_pillow"


pcolls = {}  # preview collections

classes = (
    ALPHATREE_OT_install_pillow,
    PreviewCacheItems,
)


@bpy.app.handlers.persistent
def on_load_handler(_):
    """Reload the previews when a file is loaded so that they match."""
    reload_all_previews(_, bpy.context)


def register():
    for cls in classes:
        bpy.utils.register_class(cls)

    bpy.app.handlers.load_post.append(on_load_handler)

    # register previews
    if not pcolls:
        size = 256
        pcolls["AT_tree_previews_large"] = previews.new(max_size=(size, size))
        pcolls["AT_tree_previews_small"] = previews.new(max_size=(size / 2, size / 2))
        pcolls["AT_weight_previews"] = previews.new(max_size=(size, size))

        # custom icons
        size = 32
        pcoll = previews.new(max_size=(size, size))
        path = get_resources_path(sub="icons")
        for file in os.listdir(path):
            if file.endswith(".png"):
                icon_path = os.path.join(path, file)
                pcoll.load(file, icon_path, "IMAGE")
        pcolls["AT_icons"] = pcoll


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)

    remove_handler(on_load_handler, "load_post")

    # unregister icons
    for pcoll in pcolls.values():
        previews.remove(pcoll)
    pcolls.clear()
