from dataclasses import dataclass
import bpy
import os
import webbrowser
from bpy.props import StringProperty, IntProperty, EnumProperty, BoolProperty

from .icons import get_icon
from .constants import DOCS_URL
from .functions.helper_functions import get_prefs

# flake8: noqa E501


def get_help():
    # do it lol
    help_id = get_icon("info.png")
    return help_id


def draw_info(layout: bpy.types.UILayout, title: str, info_name: str, emboss=False, text=""):
    """draw an info button in the given layout

    Args:
        layout (bpy.types.UILayout): The layout in which to draw the info button
        title (str): The title to be displayed at the top of the info window
        info_name (str): The name of the info snippet to be displayed
        emboss (bool, optional): Whether to emboss the info button
    """
    if not get_prefs().show_info:
        return
    op = layout.operator("alpha_tree.show_info", text=text, icon_value=get_help(), emboss=emboss)
    op.title = title
    op.info_name = info_name


class ALPHATREE_OT_show_info(bpy.types.Operator):
    """Show info about the selected area"""
    bl_label = "Show info"
    bl_idname = "alpha_tree.show_info"
    bl_description = "Show info about the selected area"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {
        "REGISTER",
        "UNDO",
    }

    info_name: StringProperty()

    title: StringProperty(default="Info")

    len: IntProperty(default=45)

    # docs_type: EnumProperty(
    #     items=(
    #         ("LOC", "Locally", "Open the documentation from a file on this computer"),
    #         ("WEB", "Online", "Open the documentation from from the online document (Always up to date)"),
    #     ),
    #     default="WEB",
    # )

    def invoke(self, context, event):
        return bpy.context.window_manager.invoke_popup(self)

    def draw(self, context):
        layout = self.layout
        layout.alignment = "CENTER"
        col = layout.column(align=True)
        box = col.box()
        row = box.row(align=True)
        row.alignment = "CENTER"
        row.label(text=self.title.capitalize())
        box = col.box().column(align=True)

        snippet = info_snippets[self.info_name]
        text = snippet.text
        len = self.len
        write = ""
        i = 0

        for char in text:
            if i >= len and char == " " or char == "\n":  # Write current line
                box.label(text=write.replace("  ", ""))  # Remove unnecessary spaces
                write = ""
                i = 0
            else:  # Add current character
                write += char
                i += 1

        box.label(text=write)
        box = col.box()
        row = box.row(align=True)
        # This may be useful if I make offline documentation at some point agian.
        # row.prop(self, "docs_type", text="", icon_only=True, icon="FILE_FOLDER" if self.docs_type == "LOC" else "URL")
        # row.operator("alpha_tree.open_docs", icon="URL").local = True if self.docs_type == "LOC" else False
        row.operator("alpha_tree.open_docs", icon="URL").suburl = snippet.suburl

    def execute(self, context):
        return {"FINISHED"}


class ALPHATREE_OT_open_docs(bpy.types.Operator):
    """Open documentation either online or in the addon folder"""
    bl_label = "Open docs"
    bl_idname = "alpha_tree.open_docs"
    bl_description = "Open documentation either online or in the addon folder"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER"}

    local: BoolProperty()

    suburl: StringProperty()

    def execute(self, context):
        if not self.suburl.startswith("/"):
            self.suburl = "/" + self.suburl
        bpy.ops.wm.url_open(url=DOCS_URL + self.suburl)

        # if self.local:
        #     script_path = os.path.dirname(os.path.abspath(__file__))
        #     docs_path = os.path.join(script_path, "resources", "alpha_trees_docs.pdf")
        #     path = bpy.path.abspath(docs_path)  # pylint: disable=assignment-from-no-return
        #     webbrowser.open('file:///' + path)
        # else:

        return {"FINISHED"}


classes = (
    ALPHATREE_OT_show_info,
    ALPHATREE_OT_open_docs,
)


def register():
    for cls in classes:
        bpy.utils.register_class(cls)


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)


@dataclass
class Snippet:
    text: str
    suburl: str = "/"


# To insert an intentional linebreak, don't use \ to break the line.
# Otherwise, it will be treated as the same line when drawn.

info_snippets = {
    "library_path":
        Snippet(
            """\
            这是外部库的路径，该库是保存创建的树图像，生物群落和预设的文件夹.\
            要创建新库，只需创建一个新文件夹并选择它.\
            也可以选择现有库。如果您下载了外部库，只需确保将其解压缩并选择该文件夹\
            """,
            suburl="/preferences",
        ),
    "install_pillow":
        Snippet(
            """\
            Alpha树使用一种技术来确保预览几乎是瞬间加载的 \
            当您打开树选择器，并允许高品质的图标 \
            (而不是默认的搅拌机行为，加载图标非常缓慢).
            
            但是，要使其正常工作，需要先安装一些外部库 (PIL). \
            这需要互联网连接，可能需要几分钟.

            不要担心，如果没有安装依赖关系，Alpha树仍然可以正常工作，只是慢一点\
            """,
            suburl="/preferences",
        ),
    "multi_import":
        Snippet(
            """\
            使用此方法一次导入多个树。\
            您可以使用搜索栏查找名称中包含某些单词的所有树。\
            您还可以使用这两个按钮来打开和关闭所有树，或反转选择。\
            """,
            suburl="/scattering/trees#h.e49hr9hh38pw",
        ),
    "create_lods":
        Snippet(
            """\
            这将为所选树创建不同级别的细节，一个处于中等 (四分之一)，一个处于低 (八分之一) 分辨率。\
            这些LOD对于保持低内存使用率非常有用，因为高分辨率版本可以在靠近相机的地方使用,更远的距离使用更低的分辨率.\
            """,
            suburl="/overview",
        ),
    "missing_mats":
        Snippet(
            """\
            为了渲染，一棵树必须至少有两种材质:\
            一种用于叶子，一种用于树干。\
            这是因为插件在渲染时将模型分成两部分，\
            以便能够捕获不同的层。\
            """,
            suburl="/generating-new-trees",
        ),
    "make_editable":
        Snippet(
            """\
            为了呈现，树必须附加到文件，而不是仅仅链接.\
            这是因为插件在渲染时将模型分成两部分，以便能够捕获不同的层, \
            这意味着网格需要在场景中是可编辑的，这对于链接的资源是不可能的.\
            """,
            suburl="/generating-new-trees",
        ),
    "mat_types":
        Snippet(
            """\
            这用于确定每种材料所位于的层, \
            在渲染过程中，网格将根据指定的材质及其类型进行拆分.\
            """,
            suburl="/generating-new-trees",
        ),
    "gen_settings":
        Snippet(
            """\
            - 边框填充: 最终渲染中树周围的空白空间量。如果树在图像边界之外略微裁剪，则很有用\
            - 分辨率: 最终alpha树的输出分辨率。\
            - 样本数: 用于呈现alpha树的样本数。\
            这个插件利用了降噪，结果是为了在长距离使用。 \
            这意味着，一般来说，一个高质量的alpha树可以用大约20-40个样本来渲染，而任何超过它的采样都是不需要的。\
            """,
            suburl="/generating-new-trees",
        ),
    "particle_error":
        Snippet(
            """\
            为了工作，这个插件需要能够根据材料分割树对象。\
            如果叶子是粒子，则这是不可能的，因此需要将其转换为普通网格并连接到树干\
            """,
            suburl="/generating-new-trees",
        ),
    "package_creator":
        Snippet(
            """\
            软件包是可以在计算机之间共享的zip文件，任何人都可以使用Alpha trees addon进行安装。 \
            包可以存储:\
            - 个别树木\
            - 以前出口的生物群落和生物群落集合\
            该插件将自动打包在生物群落中指定的树，即使它们尚未被选中。\
            """,
            suburl="",
        ),
    "biomes":
        Snippet(
            """\
            Biome设置是一组树，可用于组织大场景，每个树大致相当于一个普通的blender粒子系统。\
            它们是特定于场景的，这意味着您可以在多个对象上使用相同的生物群落设置。\
            """,
            suburl="/scattering/biomes",
        ),
    "biome_menu_ops":
        Snippet(
            """\
            保存生物组设置:
                这会将当前的生物群落设置保存到一个文件中，以后可以在其他场景或项目中读取该文件。\

            加载生物群系设置:\
                加载以前保存的生物群系或生物群系集合文件。\
            """,
            suburl="/scattering/biomes#h.35uva17f5uv",
        ),
    "scatter_settings":
        Snippet(
            """\
            - 种子: 此粒子槽的种子值\
            - 密度: 每平方米的树木数量\
            - 间距: 每棵树之间的最小距离\
            - 视口显示: 视口中可见粒子的百分比 \
            - 规模: 生物群落中所有树木的总体规模 \
            - 缩放随机性: 每个粒子缩放的随机量\
            - 旋转: 粒子的全局旋转 (以度为单位)\
            - 倾斜: 在Y轴上旋转树木的量\
            - 相机: 相机 (或任何对象)，使所有的树面向\
            - 使用场景摄像机: 启用后，要跟踪的摄像机将设置为场景摄像机。\
            """,
            suburl="/scattering/biomes#h.wj44n4xeqm3n",
        ),
    "wind_settings":
        Snippet(
            """\
            启用 “风” 会自动为生物群落中所有树木的旋转设置动画，使其根据噪波纹理的图案进行移动。\
            设置为:\
            - 量: 风会影响树木的因素。0为无1为全部\
            - 倾斜: 树木可以倾斜的最大角度\
            - 速度: 树木摇来的速度\
            - 随机性: 树木如何同步\
            - 图案比例: 用于控制树中的风图案的噪波纹理的大小\
            """,
            suburl="/scattering/wind",
        ),
    "dist_settings":
        Snippet(
            """\
            当您使用插件添加权重贴图时，它将自动创建一个程序权重图层堆栈，可以实时应用程序效果。\
            如果要编辑这些程序贴图，可以正常绘制，也可以在 “权重图层” 面板中编辑特定参数。\

            添加组时，可以生成各种程序权重贴图:\
            空白-这是一个正常的权重映射，其中所有值默认为0\
            高度-创建一个渐变，从对象底部的0到顶部的1\
            网格接近度-根据每个点与选定目标网格的距离创建渐变\
            点邻近性-从单个点创建渐变\
            程序纹理-可用于任何纹理，如图像，噪声，波浪，马斯格雷夫等。\
            斜率-向上指向的点的值为1，水平指向的点的值为0\
            """,
            suburl="/scattering/distribution",
        ),
    "tree_menu_ops":
        Snippet(
            """\
            多重添加:\
            这允许您一次向选定的生物群落添加多个树。\
            复制树:\
            允许您将具有特定设置的树木复制到其他生物群落。\
            """,
            suburl="/scattering/trees#h.e49hr9hh38pw",
        ),
    "tree_settings":
        Snippet(
            """\
            - 缩放: 所选树的比例\
            - 宽度: 所选树的宽度\
            - Z偏移: emiter对象上方或下方的树的高度\
            - 百分比: 要为该树显示的生物群系 “每棵树编号” 设置的百分比。\
            例如，如果 “No.per tree” 为1000，并且百分比为0，则不会显示这些树。\
            但是，如果百分比为200，则将有2000这些树.\
            """,
            suburl="/scattering/trees#h.afzo35cpz9sz",
        ),
    "material_settings":
        Snippet(
            """\
            用于此树材质的设置。\
            所有滑块从零到一，但可以通过键入更高的值来推动。\
            """,
            suburl="/scattering/material",
        ),
    "tracking_settings":
        Snippet(
            """\
            这可以用来使树总是面向某个对象。\
            """,
            suburl="/scattering/biomes#h.wj44n4xeqm3n",
        ),
    "not_current_lib":
        Snippet(
            """\
            如果在当前选定的库中找不到当前选定的树，则会发生此错误。\
            要解决此问题，请在插件首选项中选择原始库\
            """,
            suburl="",
        ),
}
