import bpy
from bpy.props import BoolProperty, StringProperty, EnumProperty

from ..functions import gen_functions, imp_functions
from ..functions.helper_functions import get_prefs

import os
from os.path import join
import time
import webbrowser
import statistics as stats

# Generator


class ALPHATREE_OT_auto_mat_type(bpy.types.Operator):
    """Manipulate list items, either: move up, move down, add or remove"""
    bl_label = "Auto detect"
    bl_idname = "alpha_tree.auto_mat_type"
    bl_description = "Automatically detect the correct type for each material"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER"}

    all: BoolProperty(default=False)

    def execute(self, context):

        at = context.scene.alpha_trees
        if not self.all:
            if at.generation_type == "BATCH":
                trees = [at.obj_list[at.obj_index].tree_obj]
            else:
                trees = [at.tree_obj]
        else:
            trees = [i.tree_obj for i in at.obj_list if i.tree_obj]

        for tree in trees:
            materials = [s.material for s in tree.material_slots]
            not_found = []

            if not materials:
                continue

            trunk_syns = ["TRUNK", "BARK", "STALK", "STEM", "LOG", "WOOD"]
            branch_syns = [
                "BRANCH",
                "SPINE",
                "PRICKLE",
                "BARB",
                "THORN",
                "SPIKE",
            ]
            leaf_syns = ["LEAF", "LEAVES", "NEEDLE", "FROND"]
            flower_syns = ["FLOWER", "BLOSSOM", "PETAL", "BLOOM"]
            fruit_syns = ["FRUIT", "BERRY", "NUT"]

            # simple search
            for mat in materials:
                if mat:
                    for type in [trunk_syns, branch_syns, leaf_syns, flower_syns, fruit_syns]:
                        for syn in type:
                            if syn in mat.name.upper():
                                mat.at.type = type[0]
                                break
                        else:
                            continue
                        break
                    else:
                        not_found.append(mat)
                else:
                    continue

            if not not_found:
                continue

            # search images
            still_not_found = []
            for mat in not_found:
                scores = [["TRUNK", 0], ["BRANCH", 0], ["LEAF", 0], ["FLOWER", 0], ["FRUIT", 0]]

                if not mat.node_tree:
                    continue
                if not mat.node_tree.nodes:
                    continue

                nodes = mat.node_tree.nodes

                for node in nodes:
                    if node.type == "TEX_IMAGE":
                        if node.image:
                            if any(w in node.image.name.upper() for w in trunk_syns):
                                scores[0][1] += 1
                            if any(w in node.image.name.upper() for w in branch_syns):
                                scores[1][1] += 1
                            if any(w in node.image.name.upper() for w in leaf_syns):
                                scores[2][1] += 1
                            if any(w in node.image.name.upper() for w in flower_syns):
                                scores[3][1] += 1
                            if any(w in node.image.name.upper() for w in fruit_syns):
                                scores[4][1] += 1

                best_match = ["TRUNK", 0]
                for score in scores:
                    if score[1] > best_match[1]:
                        best_match = score

                if best_match[1]:
                    mat.at.type = best_match[0]
                else:
                    still_not_found.append(mat)

            if not still_not_found:
                continue

            # make best guess based on nodes
            for mat in still_not_found:
                nodes = mat.node_tree.nodes
                transparent = False
                translucent = False

                for node in nodes:
                    if node.type == "BSDF_PRINCIPLED":
                        if node.inputs[18].links and node.outputs[0].links:
                            transparent = True
                    elif node.type == "BSDF_TRANSPARENT":
                        if node.outputs[0].links:
                            transparent = True
                    elif node.type == "BSDF_TRANSLUCENT":
                        if node.outputs[0].links:
                            translucent = True

                if transparent and translucent:
                    mat.at.type = "LEAF"
                elif transparent:
                    mat.at.type = "BRANCH"
                elif translucent:
                    mat.at.type = "LEAF"
                else:
                    mat.at.type = "TRUNK"

                mat.at.unsure = True

        return {"FINISHED"}


class ALPHATREE_OT_obj_list_actions(bpy.types.Operator):
    """Manipulate list items, either: move up, move down, add or remove"""
    bl_label = "List actions"
    bl_idname = "alpha_tree.obj_list_actions"
    bl_description = "Manipulate list items, either: move up, move down, add or remove"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER"}

    action: bpy.props.EnumProperty(
        items=(
            ("ADD", "Add", "Add item"),
            ("REMOVE", "Remove", "Remove item"),
            ("UP", "Up", "Move item up"),
            ("DOWN", "Down", "Move item down"),
        )
    )

    def execute(self, context):
        at = context.scene.alpha_trees
        obj_list = at.obj_list

        if self.action == "ADD":

            obj_list.add()
            new_item = obj_list[len(obj_list) - 1]
            for obj in obj_list:
                if context.object == obj.tree_obj:
                    break
            else:
                if context.object:
                    new_item.tree_obj = context.object
                    new_item.name = new_item.tree_obj.name
            at.obj_index = len(obj_list) - 1

            return {"FINISHED"}

        elif self.action == "REMOVE":

            if len(obj_list) == 0:
                return

            obj_list.remove(at.obj_index)
            at.obj_index = min(max(0, at.obj_index - 1), len(obj_list) - 1)

            return {"FINISHED"}

        elif self.action == "UP" or "DOWN":
            up = (self.action == "UP")
            length = len(obj_list) - 1
            new_index = at.obj_index + (-1 if up else 1)

            obj_list.move(new_index, at.obj_index)
            at.obj_index = max(0, min(new_index, length))

        return {"FINISHED"}


class ALPHATREE_OT_add_objs_to_list(bpy.types.Operator):
    """Add all objects in a collection as list items"""
    bl_label = "Add objects to list"
    bl_idname = "alpha_tree.add_objs_to_list"
    bl_description = "Add all objects either selected or in a collection as list items"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER"}

    type: EnumProperty(
        items=(
            ("SELECTION", "Selection", "Add the current selection as list items"),
            ("COLLECTION", "Collection", "Add all objects in a collection as list items"),
        )
    )

    def get_collection_items(self, context):
        items = [("NONE", "", "", "OUTLINER_COLLECTION", 0)]
        for i, coll in enumerate(bpy.data.collections):
            if coll.objects:
                items.append((coll.name, coll.name, coll.name, "OUTLINER_COLLECTION", i + 1))
        return items

    collection: EnumProperty(items=get_collection_items)

    def invoke(self, context, event):
        if self.type == "COLLECTION":
            return context.window_manager.invoke_props_dialog(self)
        else:
            return self.execute(context)

    def draw(self, context):
        layout = self.layout
        layout.prop(self, "collection", icon="OUTLINER_COLLECTION")

        if self.collection == "NONE":
            return

        collection = bpy.data.collections[self.collection]
        objs = context.selected_objects if self.type == "SELECTION" else collection.objects
        box = layout.box()
        box.label(text="Objects: " + str(len(objs)))

    def execute(self, context):
        at = context.scene.alpha_trees

        if self.type == "COLLECTION" and self.collection == "NONE":
            return {"FINISHED"}

        objs = context.selected_objects if self.type == "SELECTION" else bpy.data.collections[self.collection].objects
        for obj in objs:
            obj_list = at.obj_list

            for item in obj_list:
                if obj == item.tree_obj:
                    break
            else:
                bpy.ops.alpha_tree.obj_list_actions(action="ADD")  # pylint: disable=no-member
                item = obj_list[at.obj_index]
                item.tree_obj = obj

        return {"FINISHED"}


class ALPHATREE_OT_make_editable(bpy.types.Operator):
    """Manipulate list items, either: move up, move down, add or remove"""
    bl_label = "Make tree editable"
    bl_idname = "alpha_tree.make_editable"
    bl_description = "Make object link properly to scene"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER"}

    def execute(self, context):

        at = context.scene.alpha_trees
        if at.generation_type == "BATCH":
            at = at.obj_list[at.obj_index]

        try:
            gen_functions.make_tree_editable(at)
        except Exception as e:
            print(e)
            self.report(
                {"ERROR"}, "Can't make tree editable, try appending to scene if linked. Check console for details"
            )

        bpy.ops.alpha_tree.auto_mat_type()

        return {"FINISHED"}


class ALPHATREE_OT_set_up_scene(bpy.types.Operator):
    """Set up scene for rendering"""
    bl_label = "Set up scene"
    bl_idname = "alpha_tree.set_up_scene"
    bl_description = "Set up scene for rendering"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):

        at = context.scene.alpha_trees
        if at.generation_type == "BATCH":
            at = at.obj_list[at.obj_index]
        gen_functions.setup_scene(self, context, at)

        return {"FINISHED"}


class ALPHATREE_OT_set_up_tree(bpy.types.Operator):
    """Set up tree for rendering"""
    bl_label = "Set up tree"
    bl_idname = "alpha_tree.set_up_tree"
    bl_description = "Set up tree for rendering"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        at = bpy.context.scene.alpha_trees
        if at.generation_type == "BATCH":
            at = at.obj_list[at.obj_index]

        if not at.tree_obj:
            self.report({"WARNING"}, "No objects selected")
            return {"CANCELLED"}

        gen_functions.scale_to_height(at)
        gen_functions.set_render_border(context, at)
        # gen_functions.transparent_normal(self, context)

        return {"FINISHED"}


class ALPHATREE_OT_render_maps(bpy.types.Operator):
    """render all maps"""
    bl_label = "Render maps"
    bl_idname = "alpha_tree.render_maps"
    bl_description = "render all maps"
    bl_description = "render all maps"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        cscene = scene.cycles
        layout.scale_y = 1.3
        layout.prop(cscene, "samples", text="samples (50 max.)")

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    def execute(self, context):
        at = bpy.context.scene.alpha_trees
        if at.generation_type == "BATCH":
            at = at.obj_list[at.obj_index]

        if not at.tree_obj:
            self.report({"WARNING"}, "No objects selected")
            return {"CANCELLED"}

        gen_functions.render_maps(self, context, at)

        return {"FINISHED"}


class ALPHATREE_OT_open_folder(bpy.types.Operator):
    """open the renders folder"""
    bl_label = "Open renders folder"
    bl_idname = "alpha_tree.open_folder"
    bl_description = "Open renders folder"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        lib_path = os.path.join(get_prefs().lib_path, "rendered_maps", "")
        path = bpy.path.abspath(lib_path)  # pylint: disable=assignment-from-no-return
        print("Opening: " + path)
        webbrowser.open('file:///' + path)
        return {"FINISHED"}


class ALPHATREE_OT_execute_all(bpy.types.Operator):
    """Execute all operators"""
    bl_label = "Execute all"
    bl_idname = "alpha_tree.execute_all"
    bl_description = "Bake the Alpha tree maps for the selected tree"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    show_info: BoolProperty(name="Show checks", default=False)

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        cscene = scene.cycles
        layout.scale_y = 1
        layout.prop(cscene, "samples", text="samples (50 max.)")
        box = layout.box()
        row = box.row()
        row.prop(self, "show_info", text="", icon="TRIA_DOWN" if self.show_info else "TRIA_RIGHT", emboss=False)
        row.label(text="Before rendering, check:")

        if self.show_info:
            col = box.column(align=True)
            col.scale_y = 0.7
            lines = [
                "- The origin point of the model is at the base of the tree",
                "- the tree is a single mesh, with no particles or instances on it",
                "- There are no more than 50 samples (Any more are unnecessary",
                "   and add to render times)",
                "",
                "If these are not applied, the resulting images may have incorrect",
                "scaling or placement",
            ]
            imp_functions.draw_text(col, lines)

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self, width=450)

    def execute(self, context):
        at = context.scene.alpha_trees
        data = at

        if at.generation_type == "BATCH":
            start = time.time()
            times = []
            failed = []
            for item in at.obj_list:
                if not item.tree_obj or not item.show_render:
                    continue

                # for some reason, this just returns a reference to the obj location rather than a new value.
                # so it needs to be converted to a proper tuple.
                # Pretty sure this is just my bad coding but oh well.
                ob_loc = tuple(item.tree_obj.location)
                ob_scale = tuple(item.tree_obj.scale)

                data = item
                tree_start = time.time()

                # nasty catchall to prevent stopping in the middle of a large job
                try:
                    gen_functions.scale_to_height(data)
                    gen_functions.render_maps(self, context, data)

                    total_tree_time = round(time.time() - tree_start, 2)
                    print(f"\n{item.tree_obj.name} rendered in {total_tree_time} seconds")
                    times.append(total_tree_time)
                except Exception as e:
                    failed.append([item.tree_obj.name, e])
                    self.report({"ERROR"}, f"\n{item} failed. Error: {e}")

                item.tree_obj.location = ob_loc
                item.tree_obj.scale = ob_scale

            total_time = round(time.time() - start, 2)
            times.sort()
            print("\n\n\n\n")
            print(f"All {len([i for i in at.obj_list if i.show_render])} trees done in {sum(times)} seconds")
            print(f"Average: {stats.mean(times)}")
            print(f"Fastest: {times[0]}")
            print(f"Slowest: {times[-1]}")
            print(f"Failed: {len(failed)}")

            if at.open_renders_folder:
                print("Opening folder")
                bpy.ops.alpha_tree.open_folder()  # pylint: disable=no-member

            if failed:
                self.report({"ERROR"}, f"{len(failed)} trees failed. Check console for details")
                for f in failed:
                    print(f"{f[0]} failed. Error: {f[1]}")

            self.report({"INFO"}, f"All trees rendered in {sum(times)}s. Check console for details")

        else:
            if not at.tree_obj:
                self.report({"WARNING"}, "No tree object selected")
                return {"CANCELLED"}

            start = time.time()
            # for some reason, this just returns a reference to the obj location rather than a new value.
            # so it needs to be converted to a proper tuple.
            # Pretty sure this is just my bad coding but oh well.
            ob_loc = tuple(at.tree_obj.location)
            ob_scale = tuple(at.tree_obj.scale)

            gen_functions.scale_to_height(data)
            gen_functions.render_maps(self, context, data)
            if at.open_renders_folder:
                print("Opening folder")
                bpy.ops.alpha_tree.open_folder()  # pylint: disable=no-member

            total_time = round(time.time() - start, 2)
            at.tree_obj.location = ob_loc
            at.tree_obj.scale = ob_scale

            print(f"Done in {total_time} seconds")
            self.report({"INFO"}, f"Done in {total_time} seconds")

        return {"FINISHED"}


class ALPHATREE_OT_generate_lods(bpy.types.Operator):
    """Create levels of detail for a selected tree"""
    bl_label = "Create LODs"
    bl_idname = "alpha_tree.generate_lods"
    bl_description = "Create levels of detail for a selected tree"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    tree_name: StringProperty(name="Tree name", default="", options={"HIDDEN"})

    generate_all: BoolProperty(
        name="Create for all trees in library",
        description="Create LODs for all of the trees that don't have them in the current library"
    )

    do_invoke: BoolProperty(default=True)

    def invoke(self, context, event):
        if self.do_invoke:
            return context.window_manager.invoke_props_dialog(self)
        else:
            return self.execute(context)

    def draw(self, context):
        layout = self.layout
        row = layout.row()
        row.prop(self, "generate_all")

        path = join(get_prefs().lib_path, "rendered_maps")

        if self.generate_all and len(os.listdir(path)) > 60:
            layout.label(text="There are a lot of trees in this library, it could take a while")

    def execute(self, context):
        trees = []
        st = time.time()
        if not self.tree_name:
            self.tree_name = context.scene.alpha_trees.selected_tree

        if self.generate_all:
            path = join(get_prefs().lib_path, "rendered_maps")
            med_imgs = []
            if os.path.isdir(join(path, "med")):
                med_imgs = [os.listdir(join(path, "med"))]

            for img in os.listdir(path):
                if "diff-leaf.jpg" in img:
                    if img not in med_imgs:
                        trees.append(img.split("_")[0])

        else:
            trees.append(self.tree_name.split("_")[0])

        for tree in trees:
            gen_functions.generate_lods(context, tree)

        et = time.time()
        self.report({"INFO"}, f"Created LODs for {len(trees)} trees in {round(et - st, 3)}s")

        return {"FINISHED"}


classes = (
    ALPHATREE_OT_auto_mat_type,
    ALPHATREE_OT_obj_list_actions,
    ALPHATREE_OT_add_objs_to_list,
    ALPHATREE_OT_make_editable,
    ALPHATREE_OT_set_up_scene,
    ALPHATREE_OT_execute_all,
    ALPHATREE_OT_open_folder,
    ALPHATREE_OT_render_maps,
    ALPHATREE_OT_set_up_tree,
    ALPHATREE_OT_generate_lods,
)


def register():
    for cls in classes:
        bpy.utils.register_class(cls)


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)
