import os
import random
from collections.abc import Iterable
from decimal import Decimal
from os.path import join

import bpy
from bpy.props import BoolProperty, EnumProperty, FloatProperty, StringProperty

from ..constants import AT_BLEND_NAME, AT_MAT_NAME, NODE_NAMES
from ..functions import helper_functions as helpers
from ..functions import imp_functions
from ..functions.biome_functions import get_node
from ..functions.helper_functions import (
    get_at_node_group,
    get_input,
    get_mat_control_node,
    get_prefs,
    get_tree_vars,
)
from ..icons import get_icon, get_tree_previews
from ..info import draw_info
from ..weight_layers_addon.WL_functions import get_node_by_type
from ..weight_layers_addon.WL_functions import get_wl_obj as get_at_obj
from ..weight_layers_addon.WL_functions import setup_new_geo_node_tree
from .operator_utils import MultiTreeBase

# Importer


class ALPHATREE_OT_import_tree(bpy.types.Operator):
    """Import a tree from the available library"""

    bl_label = "Import Tree"
    bl_idname = "alpha_tree.import_tree"
    bl_description = "Import a tree from the available library"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    for_psystem: BoolProperty(default=False, options={"HIDDEN"})

    def execute(self, context):
        path = get_prefs().lib_path

        if self.for_psystem:
            tree = get_tree_vars(self, context)[-1]
            imp_functions.import_tree(os.path.join(path, "rendered_maps", tree.selected_tree), False, quality="high")

        else:
            at = context.scene.alpha_trees
            tree_name = at.selected_tree.split("_")[0]
            print("")
            print("Importing " + tree_name)
            tree_obj = imp_functions.import_tree(
                os.path.join(path, "rendered_maps", at.selected_tree),
                True,
                quality=at.quality,
            )
            modifier = tree_obj.modifiers.new("AT_GEO_NODE_TREE_SINGLE", "NODES")
            group = modifier.node_group
            if not group:
                group = setup_new_geo_node_tree(name="AT_GEO_NODE_TREE_SINGLE_" + tree_name)
                modifier.node_group = group
            group.name = "AT_GEO_NODE_TREE_SINGLE_" + tree_name

            tree_group = get_at_node_group(NODE_NAMES["single_tree"], type="SINGLE_TREE")
            nodes = group.nodes
            links = group.links

            in_node = get_node_by_type(group, "GROUP_INPUT")
            out_node = get_node_by_type(group, "GROUP_OUTPUT")

            tree_node = nodes.new("GeometryNodeGroup")
            tree_node.node_tree = tree_group
            tree_node.name = tree_group.name

            links.new(in_node.outputs["Geometry"], tree_node.inputs["Geometry"])
            links.new(tree_node.outputs["Geometry"], out_node.inputs["Geometry"])

            get_input(tree_node, "single_tree_obj").default_value = tree_obj

            at.import_type = "OBJECT"

        return {"FINISHED"}


class ALPHATREE_OT_reload_previews(bpy.types.Operator):
    """Reload the previews to inculde new trees"""

    bl_label = "Reload previews"
    bl_idname = "alpha_tree.reload_previews"
    bl_description = "Reload the previews to inculde new trees"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        context.scene.alpha_trees.reload_all_previews(context)
        # helpers.get_metadata(reload=True)
        return {"FINISHED"}


class ALPHATREE_OT_multi_import(MultiTreeBase):
    """Import multiple trees from the library"""

    bl_label = "Multi import"
    bl_idname = "alpha_tree.multi_import"
    bl_description = "Import multiple trees from the library"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    quality: EnumProperty(
        items=(
            ("high", "High quality", "High quality"),
            ("med", "Medium quality", "Medium quality"),
            ("low", "Low quality", "Low quality"),
        )
    )

    particle: BoolProperty(options={"HIDDEN"}) = False

    def invoke(self, context, event):
        self.quality = "med" if self.particle else "high"
        self.add_copy_items(context)
        return self.get_dialog_box(context)

    def draw(self, context):
        layout = self.layout
        row = self.draw_header(context, layout)

        row = row.row(align=True)
        draw_info(row, "Multi import", "multi_import", emboss=True)

        num = self.draw_trees(context, layout)
        if num > 10:
            row = layout.row()
            row.alignment = "CENTER"
            row.label(text="That's a lot of trees! It could take a while to import")

    def execute(self, context):
        at = context.scene.alpha_trees
        copy_items = at.copy_items

        if self.particle:
            for i in copy_items:
                if i.on:
                    bpy.ops.alpha_tree.tree_list_actions(action="ADD")  # pylint: disable=no-member
                    tree_list = get_tree_vars(self, context)[0]
                    tree = tree_list[-1]
                    tree.selected_tree = i.string
        else:
            at.quality = self.quality
            trees = []
            pos = 0
            for i in copy_items:
                if i.on:
                    at.selected_tree = i.string
                    bpy.ops.alpha_tree.import_tree()  # pylint: disable=no-member
                    trees.append(context.active_object)

            if trees == []:
                self.report({"WARNING"}, "No trees selected")
                return {"CANCELLED"}

            def get_size(object):
                return object.dimensions[2]

            trees.sort(key=get_size)
            total_len = 0
            for tree in trees:
                total_len += tree.dimensions[0]

            for tree in trees:
                pos += tree.dimensions[0] / 2
                tree.location[0] = pos - total_len / 2
                pos += tree.dimensions[0] / 2

                for i in range(2):
                    tree.location[i] += context.scene.cursor.location[i]

        return {"FINISHED"}


class ALPHATREE_OT_change_tree(bpy.types.Operator):
    """Switch to the next/previous tree"""

    bl_idname = "alpha_tree.change_tree"
    bl_label = "Next/Previous"
    bl_options = {"INTERNAL"}

    forward: bpy.props.BoolProperty()

    def execute(self, context):
        at = context.scene.alpha_trees
        diff_maps = helpers.get_preview_maps()

        if not diff_maps:
            return {"CANCELLED"}

        for i, map in enumerate(diff_maps):
            if map == at.selected_tree:
                break

        if diff_maps[i] == diff_maps[0] and not self.forward:
            at.selected_tree = diff_maps[-1]
            return {"FINISHED"}
        elif diff_maps[i] == diff_maps[-1] and self.forward:
            at.selected_tree = diff_maps[0]
            return {"FINISHED"}
        else:
            at.selected_tree = diff_maps[i + 1] if self.forward else diff_maps[i - 1]
        return {"FINISHED"}


class ALPHATREE_OT_random_tree(bpy.types.Operator):
    """Select a random tree from the library"""

    bl_label = "Select random tree"
    bl_idname = "alpha_tree.select_random_tree"
    bl_description = "Select a random tree from the library"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"

    def execute(self, context):
        at = context.scene.alpha_trees
        trees = [prev[0] for prev in get_tree_previews(at) if prev[0] != "default.png"]
        if trees:
            at.selected_tree = random.choice(trees)

        return {"FINISHED"}


class ALPHATREE_OT_edit_tree_settings(MultiTreeBase):
    """Edit the default settings of this tree"""

    bl_label = "Edit tree settings"
    bl_idname = "alpha_tree.edit_tree_settings"
    bl_description = "Edit the default settings of this tree"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"INTERNAL"}

    name: StringProperty(
        name="Name",
    )

    default_height: FloatProperty(
        name="Default scale",
        description="The default scale when imported (relative to 1m)",
    )

    split_layout: BoolProperty(
        name="Split layout",
        description="Split the layout with all trees on one side, and the selected tree on the other",
        default=True,
    )

    def invoke(self, context, event):
        copy_items = self.add_copy_items(context)
        tree = context.scene.alpha_trees.selected_tree
        for i in copy_items:
            if i.string == tree:
                i.on = True
            i.exclusive = True

        return self.get_dialog_box(context)

    def get_active_tree(self, context):
        copy_items = context.scene.alpha_trees.copy_items
        item = None
        for i in copy_items:
            if i.on:
                item = i
                break
        return item

    def draw_tree_settings(self, layout, item):
        if self.split_layout:
            super().draw_tree_settings(layout, item)
        else:
            col = layout.column(align=True)
            col.prop(item, "name", text="")
            col.prop(item, "float", text="Height (m)")
            col.prop(item, "subtype", text="")
            col.prop(item, "type", text="")
            col.prop(item, "tags", text="")

    def draw(self, context):
        layout = self.layout
        if self.split_layout:
            split = layout.split(factor=0.7)
            left = split.box()
            right = split.column(align=True)

            item = self.get_active_tree(context)

            if item is None:
                icon_val = get_icon("default.png", type="preview")
            else:
                icon_val = self.get_icon_from_copy_item(context, item, size="large")
            right.box().template_icon(icon_val, scale=self.scale * 1.75)
            box = right.box()

            if item is None:
                row = box.row()
                row.alignment = "CENTER"
                row.enabled = False
                row.label(text="            Select a tree to edit it's settings")
            else:
                box.use_property_split = True
                box.prop(item, "name")
                box.prop(item, "float", text="Height (m)")
                box.prop(item, "subtype", text="Species")
                box.prop(item, "type", text="Type")
                box.prop(item, "tags")
                box.operator("alpha_tree.delete_tree", text=f"Delete: {item.name}", icon="TRASH").tree = item.string

        else:
            left = layout

        header = self.draw_header(context, left, draw_toggles=False, draw_number=False)
        header.prop(
            self,
            "split_layout",
            toggle=True,
            text="",
            icon="ARROW_LEFTRIGHT",
        )
        self.draw_trees(context, left, scale=1.1 if self.split_layout else 1.05)

    def execute(self, context):
        copy_items = context.scene.alpha_trees.copy_items
        for item in copy_items:
            metadata = helpers.get_metadata(item.string)
            prev_name = metadata["name"]
            base_dir = join(get_prefs().lib_path, "rendered_maps")

            # Check if any changes have been made
            tags = item.tags.split(", ")
            prev_scale = round(Decimal(metadata["scale"]), 3)  # Decimal needed to avoid floating point rounding issues
            curr_scale = round(Decimal(item.float), 3)
            if (
                item.name == prev_name
                and curr_scale == prev_scale
                and item.type == metadata["type"]
                and item.subtype == metadata["subtype"]
                and tags == metadata["tags"]
            ):
                continue

            # Check that the new name does not already exist
            if item.name != prev_name:
                exists = True
                for img in os.listdir(base_dir):
                    if img.split("_")[0] == item.name:
                        self.report({"ERROR"}, f"Tree {item.name} already exists")
                        break
                else:
                    exists = False
                if exists:
                    continue

            # Rename all files including LODs
            tree_dirs = ["", "med", "low"]
            for tree_dir in tree_dirs:
                tree_dir = join(base_dir, tree_dir)
                imgs = os.listdir(tree_dir)
                for img in imgs:
                    if prev_name == img.split("_")[0] and os.path.isfile(join(tree_dir, img)):
                        prev_fpath = join(tree_dir, img)
                        img_name = img.replace(prev_name, item.name)
                        fpath = join(tree_dir, img_name)
                        os.rename(prev_fpath, fpath)

            helpers.write_metadata(
                context,
                item.name,
                scale=float(curr_scale),
                type=item.type,
                subtype=item.subtype,
                tags=tags,
            )

            # set paths to new values
            for img in bpy.data.images:
                if prev_name in img.name:
                    img_name = img.name.replace(prev_name, item.name)
                    img.filepath = join(base_dir, img_name)
                    img.name = img_name

            for mat in bpy.data.materials:
                if prev_name in mat.name:
                    mat.name = mat.name.replace(prev_name, item.name)

        bpy.ops.alpha_tree.reload_previews()
        return {"FINISHED"}


class ALPHATREE_OT_copy_settings_to_single(bpy.types.Operator):
    """Copy settings from current tree to others"""

    bl_label = ""
    bl_idname = "alpha_tree.copy_settings_to_single"
    bl_description = "Copy settings from current tree to others"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    type: bpy.props.EnumProperty(
        items=(
            ("MATERIAL", "Material", "Copy material settings"),
            ("TRACKING", "Tracking", "Copy tracking settings"),
        )
    )

    def get_materials(self, from_mat):
        mats = []
        for m in bpy.data.materials:
            if m.at.is_at and m != from_mat and m.name != AT_MAT_NAME:
                mats.append(m)
        return mats

    def get_trees(self, context):
        objs = []
        from_obj = get_at_obj(context)
        for o in bpy.data.objects:
            if o.at.is_at and o != from_obj and "AT_PARTICLE" not in o.name:
                objs.append(o)
        return objs

    def invoke(self, context, event):
        at = context.scene.alpha_trees
        copy_items = at.copy_items
        copy_items.clear()
        settings = []
        obj = get_at_obj(context)

        if self.type == "MATERIAL":
            control_node = get_mat_control_node(self, context, obj)
            mat = obj.active_material
            settings.append(
                [
                    "Quality",
                    mat.at.quality,
                    False,
                ]
            )

        elif self.type == "TRACKING":
            modifier = [m for m in obj.modifiers if m.type == "NODES"][0]
            control_node = get_node(modifier.node_group, "single_tree")

            settings.append(
                [
                    "Use scene camera",
                    obj.at.use_scene_cam,
                    False,
                ]
            )

        settings.extend(
            [
                [i.name, i.default_value, i.hide_value]
                for i in control_node.inputs
                if i.type != "GEOMETRY" and "internal_" not in i.name
            ]
        )

        for setting in settings:
            copy_items.add()
            new_item = copy_items[len(copy_items) - 1]

            new_item.name = setting[0]

            if setting[2]:
                new_item.hide = True
                new_item.on = False
            else:
                if isinstance(setting[1], str):
                    new_item.string = setting[1]
                elif isinstance(setting[1], bool):
                    new_item.bool = setting[1]
                elif isinstance(setting[1], bpy.types.Object):
                    new_item.object = setting[1]
                elif isinstance(setting[1], Iterable):  # is color
                    new_item.color = setting[1]
                elif isinstance(setting[1], float):
                    new_item.float = setting[1]

        return context.window_manager.invoke_props_dialog(self, width=450)

    def draw(self, context):
        layout = self.layout
        type = self.type
        at = context.scene.alpha_trees
        copy_items = at.copy_items

        split = layout.split()
        row = split.row(align=True)
        row.label(text="Copy settings:")
        row = row.row(align=True)
        row.active = False
        op = row.operator(
            "alpha_tree.toggle_copy_items",
            text="",
            emboss=False,
            icon_value=get_icon("toggle.png"),
        )
        op.type = "TOGGLE"
        op.data = ""
        op = row.operator(
            "alpha_tree.toggle_copy_items",
            emboss=False,
            text="",
            icon_value=get_icon("invert.png"),
        )
        op.type = "INVERT"
        op.data = ""

        row = split.row(align=True)
        row.label(text="Copy to:")

        row = row.row(align=True)
        row.active = False
        from_mat = get_at_obj(context).at.material
        mats = self.get_materials(from_mat)
        trees = self.get_trees(context)
        toggle_data = [mat.at for mat in mats] if self.type == "MATERIAL" else [tree.at for tree in trees]

        op = row.operator(
            "alpha_tree.toggle_copy_items",
            text="",
            emboss=False,
            icon_value=get_icon("toggle.png"),
        )
        op.type = "TOGGLE"
        op.data = str(toggle_data)
        op = row.operator(
            "alpha_tree.toggle_copy_items",
            emboss=False,
            text="",
            icon_value=get_icon("invert.png"),
        )
        op.type = "INVERT"
        op.data = str(toggle_data)

        row = layout.row()
        boxcol = row.box().column(align=True)
        for item in copy_items:
            if item.hide:
                textrow = boxcol.row()
                textrow.label(text=item.name)
            else:
                boxcol.prop(item, "on", text=item.name.split("_")[-1], toggle=True, icon_value=item.icon)

        col = row.column(align=True)
        boxcol = col.box().column(align=True)

        if type == "MATERIAL":
            if not mats:
                boxcol.enabled = False
                boxcol.label(text="No tree materials to copy to...")
                return

            boxcol.label(text="Tree materials:")

            for mat in mats:
                custom_icon_val = mat.at.icon_value
                boxcol.prop(mat.at, "copy_to", text=mat.name, icon_value=custom_icon_val)

        elif type == "TRACKING":
            if not trees:
                boxcol.enabled = False
                boxcol.label(text="No tree objects to copy to...")
                return

            for tree in trees:
                boxcol.prop(tree.at, "copy_to", text=tree.name, toggle=True)

        return

    def execute(self, context):
        at = context.scene.alpha_trees
        copy_items = at.copy_items
        type = self.type
        recps = 0

        def copy_node_settings(copy_items, from_node, to_node):
            for copy_item, from_socket in zip(copy_items, from_node.inputs):
                if not copy_item.on or from_socket.name.startswith("internal_") or from_socket.type == "GEOMETRY":
                    continue
                try:
                    to_socket = to_node.inputs[from_socket.name]
                except KeyError:
                    self.report({"ERROR"}, "Materials do not mathch, could not copy " + from_socket.name)
                    continue
                to_socket.default_value = from_socket.default_value

        if self.type == "MATERIAL":
            from_mat = get_at_obj(context).at.material
            control_node = get_node(from_mat.node_tree, "mat_control")
            mats = self.get_materials(from_mat)

            for to_material in mats:
                if to_material.at.copy_to and to_material != from_mat:
                    recps += 1
                    if copy_items[0].on:
                        to_material.at.quality = copy_items[0].string
                    to_node = get_node(to_material.node_tree, "mat_control")
                    copy_node_settings(copy_items[1:], control_node, to_node)

        elif type == "TRACKING":
            trees = self.get_trees(context)
            from_mod_group = [m for m in get_at_obj(context).modifiers if m.type == "NODES"][0].node_group
            from_node = get_node(from_mod_group, "single_tree")

            for tree in trees:
                if tree.at.copy_to:
                    recps += 1
                    if copy_items[0].on:
                        tree.at.use_scene_cam = copy_items[0].bool
                    to_mod_group = [m for m in tree.modifiers if m.type == "NODES"][0].node_group
                    to_node = get_node(to_mod_group, "single_tree")
                    copy_node_settings(copy_items[1:], from_node, to_node)

            # This is a lot more complex than it could be,
            # this is to account for differences between node groups
            # if they are edited or updated in future versions

            # for mat in mats:
            #     if mat.at.copy_to and not mat == from_mat:
            #         recps += 1
            #         groups = {}
            #         cur_group = []
            #         name = ""

            #         # get list of groups and their inputs
            #         for i, input in enumerate(src_inputs):

            #             if input.hide_value or i == len(control_node.inputs) - 1:

            #                 # check if last input
            #                 if i == len(control_node.inputs) - 1:
            #                     cur_group.append([input, copy_items[i]])

            #                 if cur_group:
            #                     groups[name] = cur_group
            #                     cur_group = []

            #                 name = input.name

            #             else:
            #                 cur_group.append([input, copy_items[i]])

            #         # copy to. I was running out of variable names here, which is probably a bad sign.
            #         control_node = get_node(mat.node_tree, "mat_control")
            #         for g in groups:

            #             for i, input in enumerate(control_node.inputs):
            #                 if input.name == g:

            #                     new_group_items = []
            #                     for j in control_node.inputs[i + 1:]:
            #                         if j.hide_value:
            #                             break
            #                         new_group_items.append(j)

            #                     for input1 in groups[g]:
            #                         if input1[1].on:
            #                             input1 = input1[0]
            #                             for input2 in new_group_items:
            #                                 if input1.name == input2.name:
            #                                     input2.default_value = input1.default_value
            #                                     break
            #                             else:
            #                                 self.report({"ERROR"},
            #                                             f"Materials don't match, could not copy '{input1.name}'")

        on = len([i for i in copy_items if i.on])
        s1 = "y" if on == 1 else "ies"
        s2 = "" if recps == 1 else "s"
        self.report({"INFO"}, f"Copied {str(on)} propert{s1} to {str(recps)} recipient{s2}")

        return {"FINISHED"}


class ALPHATREE_OT_new_at_material(bpy.types.Operator):
    """Add a new alpha trees material"""

    bl_label = "New alpha material"
    bl_idname = "alpha_tree.new_at_material"
    bl_description = "Add a new alpha trees material"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        image_path = join(
            bpy.path.abspath(get_prefs().lib_path), "rendered_maps", context.scene.alpha_trees.selected_tree
        )
        preview_name = os.path.basename(image_path)
        tree_name = preview_name.split("_")[0]

        for material in bpy.data.materials:
            if material.name == AT_MAT_NAME:
                base_tree_material = material
                break
        else:
            script_path = os.path.dirname(os.path.abspath(__file__))
            blend_path = os.path.join(script_path, "resources", AT_BLEND_NAME)
            base_tree_material = imp_functions.import_material(blend_path)

        tree_material = base_tree_material.copy()
        tree_material.name = "AT_" + tree_name

        obj = get_at_obj(context)
        mat_slot = obj.material_slots[0]
        mat_slot.material = tree_material
        tree_material.at.is_at = True
        obj.at.material = tree_material

        quality = context.scene.alpha_trees.quality

        # Switch images
        images = imp_functions.get_other_maps(context, preview_name, quality)
        labels = [
            NODE_NAMES["leaf_col"],
            NODE_NAMES["trunk_col"],
            NODE_NAMES["leaf_nor"],
            NODE_NAMES["trunk_nor"],
            NODE_NAMES["mask"],
        ]

        nodes = tree_material.node_tree.nodes
        for node in nodes:
            if node.type == "TEX_IMAGE":
                for label, img in zip(labels, images):
                    if node.label == label:
                        node.image = img

        # scale to correct ratio
        verts = obj.data.vertices
        if len(verts) == 4:
            ratio = img.size[1] / img.size[0]

            verts[2].co = (-1, 0, 2 * ratio)
            verts[3].co = (1, 0, 2 * ratio)

        return {"FINISHED"}


class ALPHATREE_OT_remove_at_material(bpy.types.Operator):
    """Delete an alpha trees material"""

    bl_label = "Delete alpha material"
    bl_idname = "alpha_tree.remove_at_material"
    bl_description = "Delete the selected alpha trees material"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        obj_at = get_at_obj(context).at
        mat = bpy.data.materials[obj_at.material.name]

        nodes = mat.node_tree.nodes
        images = []

        for node in nodes:
            if node.type == "TEX_IMAGE":
                if node.image:
                    if node.image.users <= 1:
                        images.append(node.image)

        for img in images:
            bpy.data.images.remove(img)

        bpy.data.materials.remove(mat)

        return {"FINISHED"}


class ALPHATREE_OT_delete_tree(bpy.types.Operator):
    """Delete this tree from the library"""

    bl_label = "Delete tree"
    bl_idname = "alpha_tree.delete_tree"
    bl_description = "Delete this tree from the library"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"INTERNAL"}

    tree: StringProperty()

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        layout.label(text="Are you sure you want to delete this tree permenantly?")

    def execute(self, context):
        tree = self.tree
        tree_dir = join(get_prefs().lib_path, "rendered_maps")

        for folder in ["", "med", "low"]:
            img_dir = join(tree_dir, folder)
            for img in os.listdir(img_dir):
                if tree.split("_")[0] == img.split("_")[0]:
                    os.remove(join(img_dir, img))

        context.scene.alpha_trees.selected_tree = "default.png"

        bpy.ops.alpha_tree.reload_previews()
        return {"FINISHED"}


classes = (
    ALPHATREE_OT_import_tree,
    ALPHATREE_OT_change_tree,
    ALPHATREE_OT_random_tree,
    ALPHATREE_OT_edit_tree_settings,
    ALPHATREE_OT_multi_import,
    ALPHATREE_OT_reload_previews,
    ALPHATREE_OT_copy_settings_to_single,
    ALPHATREE_OT_new_at_material,
    ALPHATREE_OT_remove_at_material,
    ALPHATREE_OT_delete_tree,
)


def register():
    for cls in classes:
        bpy.utils.register_class(cls)


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)
