import bpy
from bpy.props import *

from ..functions.helper_functions import get_metadata, get_preview_maps
from ..icons import get_icon, PreviewSettingsParent


class MultiTreeBase(bpy.types.Operator, PreviewSettingsParent):
    """Used for displaying a popup with all trees displayed"""

    tree_search: StringProperty(
        name="",
        default="",
        description="Show trees containing these words",
    )

    scale: FloatProperty(name="scale", description="Scale of icons", default=6, min=3, max=15)

    show: BoolProperty(name="Show all", default=True)

    selected_no = 0

    def add_copy_items(self, context):
        """Add a copy item for every tree in the current library."""
        at = context.scene.alpha_trees
        copy_items = at.copy_items
        copy_items.clear()
        self.show = True

        for prev in get_preview_maps():
            metadata = get_metadata(prev)
            copy_items.add()
            new_item = copy_items[len(copy_items) - 1]
            new_item.on = False
            new_item.string = prev
            new_item.name = metadata["name"]
            new_item.float = metadata["scale"]
            new_item.type = metadata["type"]
            new_item.subtype = metadata["subtype"]
            new_item.tags = ", ".join(metadata["tags"])

        return copy_items

    def get_dialog_box(self, context):
        """Return a dialog box with a suitible width for the current window and library"""
        # gives a window that scales depending on library size and window size, with a hard upper limit
        trees = context.scene.alpha_trees.copy_items
        width = min(context.window.width - (context.window.width / 4), max(len(trees) * 40, 500), 1500)
        return context.window_manager.invoke_props_dialog(self, width=int(width))

    def draw_header(self, context, layout, draw_toggles=True, draw_number=True):
        "Draw the header with the search bar, scale, toggles etc."

        big_row = layout.row(align=True)
        if draw_number:
            big_row = big_row.split(factor=0.98)
        self.header_row = big_row
        row1 = big_row.row()
        row1.prop(self, "show", text="", icon="TRIA_DOWN" if self.show else "TRIA_RIGHT")
        split = row1.split(factor=0.5334)
        split.prop(self, "tree_search", icon="VIEWZOOM")
        split.prop(self, "scale")

        if hasattr(self, "quality"):
            split.prop(self, "quality", text="")

        if draw_toggles:
            row = row1.row(align=True)
            op = row.operator("alpha_tree.toggle_copy_items", text="", icon_value=get_icon("toggle.png"))
            op.key = self.tree_search
            op.type = "TOGGLE"
            op = row.operator("alpha_tree.toggle_copy_items", text="", icon_value=get_icon("invert.png"))
            op.key = self.tree_search
            op.type = "INVERT"

        if draw_number:
            big_row.label(text=str(self.selected_no))
        return row1

    def draw_tree_settings(self, layout, item):
        """Draw the settings for the given tree"""
        row = layout.row()
        row.scale_y = self.scale / 4
        row.prop(item, "on", text=item.name.replace("-", " "), toggle=True)
        return

    def get_icon_from_copy_item(self, context, copy_item, size="small"):
        """Gets the tree preview from the name given in copy_item.string"""
        icon_value = get_icon(copy_item.string, type="preview", size=size, data=self)
        return icon_value

    def draw_trees(self, context, layout, scale=1):
        """Draw a grid of trees"""
        copy_items = context.scene.alpha_trees.copy_items

        grid = layout.grid_flow(row_major=True, align=True, columns=0)
        search_keys = self.tree_search.lower().replace(",", " ").split(" ")

        if not self.show:
            return 0

        found = 0
        num = 0
        for i in copy_items:
            tags = [t.lower() for t in i.tags.split(", ")]

            match = False
            for key in search_keys:
                if key in i.string.lower() or key in i.type.lower() or key in tags:
                    match = True

            if not search_keys[0] or match:
                
                icon_value = self.get_icon_from_copy_item(context, i)

                box = grid.box()
                col1 = box.column(align=True)
                col = col1.column(align=True)
                col.template_icon(icon_value=icon_value, scale=self.scale * scale)
                col.scale_y = 1 - (scale - 1) * 3
                self.draw_tree_settings(col1, i)
                found += 1
            if i.on:
                num += 1

        self.selected_no = num
        if not found:
            row = layout.row()
            row.alignment = "CENTER"
            row.alert = True
            row.label(text="No trees matching that search")
        return num
