import bpy
import os

from bpy.props import BoolProperty, StringProperty
from os.path import join as osjoin

from .icons import get_icon, get_tree_previews

from .t3dn_bip.utils import support_pillow
from .info import draw_info
from .functions.helper_functions import get_metadata, is_float, write_metadata
from .weight_layers_addon.WL_preferences import WeightLayersPrefsParent
from .ui import ALPHATREE_PT_import_panel, ALPHATREE_PT_settings_panel, ALPHATREE_PT_gen_panel
from .constants import VARIANT


def lib_path_update(self, context):
    if self.lib_path.startswith("//"):
        self.lib_path = bpy.path.abspath(self.lib_path)
        return

    if not self.lib_path:
        return

    if not os.access(self.lib_path, os.W_OK):
        return

    path = self.lib_path

    if os.path.isdir(path):
        for file in os.listdir(path):
            if file == "rendered_maps":
                rmaps_dir = osjoin(path, file)

                # update previous libraries to use metadata json files
                for file in os.listdir(rmaps_dir):
                    elements = file.split("_")
                    if "preview.png" in file and not os.path.exists(osjoin(rmaps_dir, elements[0] + "_data.json")):
                        # print(file)
                        type = " ".join(file.split(" ")[:-1])

                        if is_float(elements[1]):
                            scale = float(elements[1]) * 10
                        else:
                            scale = 10
                        write_metadata(context, file, scale, type, [])
                break
        else:
            split = os.path.split(path)
            if split[0].endswith("rendered_maps") or split[0].endswith("biomes"):
                self.lib_path = osjoin(os.path.split(split[0])[0], "")
            else:
                os.mkdir(osjoin(path, "rendered_maps"))
                os.mkdir(osjoin(path, "biomes"))

    else:
        dirname = os.path.dirname(path)
        if dirname.endswith("rendered_maps") or dirname.endswith("biomes"):
            split = os.path.split(dirname)
            self.lib_path = osjoin(split[0], "")
        else:
            os.mkdir(osjoin(dirname, "rendered_maps"))
            os.mkdir(osjoin(path, "biomes"))

    get_metadata(reload=True)
    # update tree preview cache for single tree importing.
    get_tree_previews(data=context.scene.alpha_trees, reload=True)
    # update tree preview caches for all existing trees in this file.
    try:
        for ng in bpy.data.node_groups:
            if ng.type == "GEOMETRY":
                for tree in ng.at.tree_list:
                    print(tree)
                    get_tree_previews(data=tree, reload=True)
    except AttributeError:
        # if tries to run when registering and restrict data is active.
        pass
    context.scene.alpha_trees.selected_tree = "default.png"


def category_update(self, context):
    panels = [
        ALPHATREE_PT_import_panel,
        ALPHATREE_PT_settings_panel,
        ALPHATREE_PT_gen_panel,
    ]
    for panel in panels:
        try:
            bpy.utils.unregister_class(panel)
        except RuntimeError:
            # Not yet registered
            pass
        panel.bl_category = self.category
        bpy.utils.register_class(panel)


class AlphaTreesPrefs(WeightLayersPrefsParent, bpy.types.AddonPreferences):
    bl_idname = __package__

    def check_lib_path(self):
        if not self.lib_path:
            return "Please select a directory for the generated trees to be saved to or, select an existing library:"

        if not os.path.isdir(self.lib_path):
            return "This library folder does not exist. Please select another one."

        elif not os.access(self.lib_path, os.W_OK):
            return "Cannot write to this folder. Please either change permissions of folder,\
                or select a different folder".replace("  ", "")
        return ""

    prefs_mode: bpy.props.EnumProperty(
        items=(
            ("BASIC", "Basic", "Show basic settings"),
            ("ADVANCED", "Advanced", "Show advanced settings"),
            ("WEIGHT_LAYERS", "Weight Layers", "Show weight layers settings"),
        ),
        name="Display mode",
        description="The type of settings to display",
        default="BASIC",
    )

    show_dev_options: BoolProperty(
        name="Dev options",
        description="Show extra operators and settings for debugging",
        default=False,
    )

    show_info: BoolProperty(
        name="Dev options",
        description="Show info buttons around the addon",
        default=True,
    )

    show_botaniq_operator: BoolProperty(
        name="Show botaniq operator",
        description="Show botaniq plant tree button if botaniq is installed for easy access",
        default=True,
    )

    botaniq_version: bpy.props.EnumProperty(
        items=(
            ("OLD", "Before v6.0", "This impacts some data paths"),
            ("NEW", "After v6.0", "This impacts some data paths"),
        ),
        name="Botaniq version",
        description="The botaniq version that is currently installed",
        default="NEW",
    )

    lib_path: StringProperty(
        name="",
        default="",
        description="Directory to save rendered images to",
        subtype="DIR_PATH",
        update=lib_path_update,
        options={"TEXTEDIT_UPDATE"},
    )

    show_package_creator: BoolProperty(
        name="Show package creator panel",
        description="Show a panel for creating packages to share with other users",
        default=False,
    )

    use_mat_icons: BoolProperty(
        name="Use material previews as tree icons",
        description="use them instead of baked previews. this can result in some squashing of icons",
        default=True,
    )

    use_arrows: BoolProperty(
        name="Use arrows for opening panels instead of icons",
        description="This uses arrows instead of different icons for each panel, which can increase clarity",
        default=False,
    )

    auto_pack_images: BoolProperty(
        name="Auto pack images",
        description="This will automatically pack tree images into the blend file. This increases file size,\
        but also prevents textures being lost if the blend or tree library is moved".replace("  ", ""),
        default=True,
    )

    category: StringProperty(
        name="N-Panel category",
        default="Alpha Trees",
        description="The category to put all of the addon panels in",
        update=category_update,
    )

    increase_bounces: BoolProperty(
        name="Increase transparency bounces when adding particles",
        description="This prevents black boxes from showing up where two many layers of alphas show up",
        default=True,
    )

    export_dir: StringProperty(
        name="Export directory",
        description="Directory to export packages to",
        default="//",
        subtype="DIR_PATH",
    )

    def draw(self, context):
        layout = self.layout  # pylint: disable=no-member
        col = layout.column().box()

        if bpy.app.version < (3, 0, 0):
            col.label(text="This addon requires Blender verison 3.0 or later to work :(")
            return

        if not self.lib_path:
            col.alert = not bool(self.lib_path)
            col.label(
                text="Please select a directory for the generated trees to be saved to or, select an existing library:")
        else:
            row = col.row()
            row.alignment = "CENTER"
            row.label(text="            Library directory:")

        row = col.row(align=True)
        draw_info(row, "External libraries", "library_path", emboss=True)
        row.prop(self, "lib_path")

        if not self.lib_path:
            return

        message = self.check_lib_path()
        if message:
            col.alert = True
            col.label(text=message)

        box = layout.box()
        col = box.column()
        row = col.row()
        row.alignment = "CENTER"
        row.label(text="            Fast loading previews:")

        pillow_installed = support_pillow()
        row = col.row(align=True)
        row.scale_y = 1.3
        row.scale_x = 1.15
        draw_info(row, "Install dependancies", "install_pillow", emboss=not pillow_installed)

        if pillow_installed:
            row.active = False
            row = row.row()
            row.enabled = False
            row.operator("alpha_tree.install_pillow", emboss=False, text="All dependancies successfully installed!")
        else:
            row.operator(
                "alpha_tree.install_pillow",
                icon="IMPORT",
            )

        col = layout.column(align=True)
        box = col.box()
        row = box.row()
        row.alignment = "CENTER"
        row.label(text="     Settings:")

        row = col.row(align=True)
        row.prop(self, "prefs_mode", expand=True)

        box = col.box()

        if self.prefs_mode == "BASIC":
            box.prop(self, "auto_pack_images")
            box.prop(self, "show_info", text="Show info buttons")
            # box.prop(self, "show_package_creator")

            for addon in context.preferences.addons:
                if "botaniq" in addon.module:
                    box.prop(self, "show_botaniq_operator")
                    if self.show_botaniq_operator:
                        box.prop(self, "botaniq_version")

        elif self.prefs_mode == "ADVANCED":
            box.prop(self, "show_dev_options", text="Show debug options")
            box.prop(self, "use_mat_icons")
            if VARIANT != "LITE":
                box.prop(self, "increase_bounces")
            box.prop(self, "use_arrows")
            box.prop(self, "category", text="N-Panel category")
        elif self.prefs_mode == "WEIGHT_LAYERS":
            box.separator()
            self.draw_wl_prefs(box, context)

        # Contacts
        box = layout.box()
        row = box.row()
        row.alignment = "CENTER"
        row.label(text="Contact:")
        col = box.column(align=True)
        row = col.row(align=True)
        op = row.operator("wm.url_open", text="Blender Artists", icon_value=get_icon("blender_artists.png"))
        op.url = "https://blenderartists.org/t/alpha-trees-render-massive-forests-fast/1250694"
        op = row.operator("wm.url_open", text="Discord", icon_value=get_icon("discord.png"))
        op.url = "https://discord.com/invite/5cVzT4pVrM"
        row = col.row(align=True)
        op = row.operator("wm.url_open", text="Blender Market", icon_value=get_icon("blender_market.png"))
        op.url = "https://blendermarket.com/creators/strike-digital"
        op = row.operator("wm.url_open", text="Twitter", icon_value=get_icon("twitter.png"))
        op.url = "https://twitter.com/StrikeDigital1"


def register():
    bpy.utils.register_class(AlphaTreesPrefs)


def unregister():
    bpy.utils.unregister_class(AlphaTreesPrefs)
