import bpy
from bpy.props import (
    BoolProperty,
    CollectionProperty,
    EnumProperty,
    FloatProperty,
    FloatVectorProperty,
    IntProperty,
    IntVectorProperty,
    PointerProperty,
    StringProperty,
)

from . import icons
from .constants import IS_DEV, VARIANT
from .functions import gen_functions, imp_functions
from .functions.callbacks import *
from .weight_layers_addon.extra_props import GetSetProperty, NonIDProperty

# property groups

tree_prev_dict = dict()


class ATGeneratorSettings(icons.PreviewSettingsParent):
    reload_all_previews = icons.reload_all_previews

    # Show UI
    # ----------------------------------------------------------

    import_type: EnumProperty(
        items=(
            (
                "PARTICLE",
                "Particle",
                "Import trees into ready to use particle systems with easy settings",
                "PARTICLE_DATA",
                1,
            ),
            ("OBJECT", "Object", "Import trees to the viewport directly for use by the use", "OBJECT_DATA", 0),
        ),
        default="OBJECT" if VARIANT == "LITE" else "PARTICLE",
    )

    generation_type: EnumProperty(
        items=(
            ("SINGLE", "Single", "Bake the maps for a single tree at a time", "OBJECT_DATAMODE", 0),
            ("BATCH", "Batch", "Bake maps for multiple trees at once", "OUTLINER_COLLECTION", 1),
        ),
        default="SINGLE",
    )

    show_track_settings: BoolProperty(
        name="settings",
        description="Show Tracking settings",
        default=False,
    )

    show_wind_settings: BoolProperty(
        name="settings",
        description="Show wind settings",
        default=False,
    )

    show_mat_settings: BoolProperty(
        name="settings",
        description="Show material settings",
        default=False,
    )

    show_gen_settings: BoolProperty(
        name="settings",
        description="Show generator settings",
        default=False,
    )

    show_vgroup_selector: BoolProperty(description="Show vertex group selector", default=True)

    show_weight_layers: BoolProperty(description="Show weight layers", default=True)

    # generator operator settings
    # ----------------------------------------------------------

    tree_obj: PointerProperty(
        type=bpy.types.Object,
        name="Tree",
        description="Tree to generate the maps for",
        poll=gen_functions.tree_obj_poll,
        update=gen_functions.tree_obj_update,
    )

    obj_index: IntProperty(default=0)

    border_padding: FloatProperty(
        name="Render padding",
        default=0.03,
        description="Amount of extra space around object in the final image\
            (Having more can slow down the rendering of alpha trees when imported slightly)",
        soft_min=0,
        soft_max=0.2,
    )

    resolution: IntProperty(name="Resolution", default=1024, description="Max resolution of images", subtype="PIXEL")

    diff_render: BoolProperty(
        name="render diffuse map",
        default=True,
        description="render the diffuse map",
    )

    nor_render: BoolProperty(
        name="render normal map",
        default=True,
        description="render the normal map",
    )

    mask_render: BoolProperty(
        name="render mask map",
        default=True,
        description="render the mask map",
    )

    preview_render: BoolProperty(
        name="render preview",
        default=True,
        description="render a preview map",
    )

    overwrite: BoolProperty(
        name="Overwrite",
        default=True,
        description="Overwrite images if they already exist",
    )

    remove_extra_masks: BoolProperty(
        name="delete extra masks",
        default=True,
        description="delete extra masks after combining (turn off for debug)",
    )

    open_renders_folder: BoolProperty(
        name="open renders folder on completion",
        default=False,
        description="open the renders folder on completion",
    )

    # internal
    auto_material_update: BoolProperty(
        default=True,
        description="Turns off the auto material selection when changing the tree_obj.\
            Used when changing it while generating a tree to retain user defined material types",
    )

    # importer operator settings
    # ----------------------------------------------------------

    selected_tree: EnumProperty(
        items=icons.get_tree_enum_previews,
        update=single_selected_tree_update,
    )

    quality: EnumProperty(
        items=imp_functions.get_tree_quality_items,
        name="Quality",
        description="The resolution to import this tree as",
    )

    update_selected_tree: BoolProperty(
        default=True,
    )

    show_import_settings: BoolProperty(
        name="Show import settings",
        description="Show import settings",
        default=False,
    )

    # show importer settings
    # ----------------------------------------------------------

    show_scatter_settings: BoolProperty(
        name="Show settings",
        default=True,
        description="Show settings",
    )

    show_tree_settings: BoolProperty(
        name="Show settings",
        default=True,
        description="Show settings",
    )

    show_material_settings: BoolProperty(
        name="Show settings",
        default=True,
        description="Show settings",
    )

    show_weight_settings: BoolProperty(
        name="Show settings",
        default=False,
        description="Show settings",
    )


class ATMaterialSettings(bpy.types.PropertyGroup):
    type: EnumProperty(
        items=gen_functions.mat_type_items,
        name="Material type",
        description="This determines what layer each part of the tree should be put on",
        update=gen_functions.mat_type_update,
    )

    name: StringProperty()

    icon_value: IntProperty(get=get_material_icon_value)

    is_at: BoolProperty(
        description="Whether this material was created by Alpha Trees",
        default=False,
    )

    quality: EnumProperty(
        items=imp_functions.get_tree_quality_items,
        name="Quality",
        description="The resolution of this tree as",
        update=quality_update,
    )

    unsure: BoolProperty(
        description="Whether the addon could automatically find the correct type for this material",
        default=False,
    )

    copy_to: BoolProperty(
        description="Whether Whether to copy settings to this material or not.",
    )

    version: IntVectorProperty(
        description="The Alpha Trees version that this material was created with.",
        default=(0, 0, 0),
    )


class ATObjectSettings(bpy.types.PropertyGroup):
    is_at: BoolProperty(
        default=False,
    )

    version: IntVectorProperty(default=(0, 0, 0))

    use_scene_cam: BoolProperty(
        name="Use scene camera",
        default=False,
    )

    material: PointerProperty(
        type=bpy.types.Material,
        name="Material",
        description="The alpha material to use for this tree",
        update=imp_functions.material_update,
        poll=imp_functions.material_poll,
    )

    camera: PointerProperty(
        type=bpy.types.Object,
        name="Camera",
        description="Camera to track trees to",
        poll=imp_functions.camera_poll,
        update=imp_functions.camera_update,
    )

    rotation_offset: FloatProperty(
        name="Offset",
        description="Amount to offset the rotation of the tree",
        subtype="ANGLE",
        default=0,
        update=imp_functions.rotation_offset_update,
    )

    invert_track: BoolProperty(
        name="Invert direction",
        description="Invert which side is facing the camera",
        default=False,
        update=imp_functions.invert_track_update,
    )

    # internal

    b_index: IntProperty(
        default=0,
        update=biome_index_update,
    )

    copy_to: BoolProperty()

    prev_camera: PointerProperty(
        type=bpy.types.Object,
    )

    # functions

    new_biome_slot = new_biome_slot

    def get_mesh_area(self, use_modifiers: bool = False) -> float:
        """Gets the total area of the current object

        Args:
            use_modifiers (bool, optional): Whether to evaluate modifiers (slower). Defaults to False.
        """
        return get_obj_area(self.id_data, use_modifiers)


class ATNodeGroupSettings(bpy.types.PropertyGroup):
    # internal
    type: EnumProperty(
        items=(
            ("None", "None", "None"),
            ("SETTINGS", "Settings", "Settings"),
            ("DISTRIBUTION", "Distribution", "Distribution"),
        ),
    )

    t_index: IntProperty(
        default=0,
    )

    prev_camera: PointerProperty(
        type=bpy.types.Object,
    )

    version: IntVectorProperty(
        description="The addon version that this group was created with",
        default=(0, 0, 0),
    )

    copy_to: BoolProperty(
        description="Whether Whether to copy settings to these biome settings or not.",
    )

    # exposed

    name: StringProperty(
        name="Name",
        description="The name of these biome settings",
        default="Biome Settings 99",
        update=settings_name_update,
    )

    use_scene_cam: BoolProperty(
        name="Use scene camera",
        description="Use the current scene camera. this updates every frame",
        default=False,
        update=imp_functions.use_scene_cam_update,
    )


class DistSettingsParent(bpy.types.PropertyGroup):
    def _get_biome_slot(self):
        for biome_slot in self.id_data.at.biome_list:
            type = biome_slot.density if self.density else biome_slot.scale
            if type == self:
                return biome_slot

    biome_slot = GetSetProperty(get=_get_biome_slot)

    vgroup_enum: EnumProperty(
        items=get_vgroup_enum_items,
        name="Selected vertex group",
        description="Choose the vertex group to control the density/scale",
        update=vgroup_enum_update,
    )

    stack_slot_name: StringProperty(description="The unique name of the layer stack slot that this references")

    # The backend reference to the vertex group used for density/scale. Uses NonIDProps:
    # https://github.com/strike-digital/non_id_props
    vgroup = NonIDProperty(
        name="vgroup",
        subtype="vertex_groups",
        update=vgroup_update,
        name_update=vgroup_update,
    )

    vgroup_name: StringProperty(
        get=vgroup_name_get,
        set=vgroup_name_set,
    )

    wlayer_slot = GetSetProperty(
        get=wlayer_slot_get,
        set=wlayer_slot_set,
    )

    # this is a shortcut to the layer stack's vgroup
    # vgroup = GetSetProperty(get=lambda self: self.wlayer_stack.vgroup if self.wlayer_stack else None)

    new_vgroup = new_vgroup

    remove_vgroup = remove_vgroup


class ATDensitySettings(DistSettingsParent):
    density = True


class ATScaleSettings(DistSettingsParent):
    density = False


# collection properties


class ObjectListItem(bpy.types.PropertyGroup):
    # display

    name: StringProperty(name="Name", description="Tree display name", default="Untitled")

    show_render: BoolProperty(
        name="Bake tree",
        default=True,
        description="Bake tree",
    )

    tree_obj: PointerProperty(
        type=bpy.types.Object,
        name="Tree",
        description="Tree to generate the maps for",
        poll=gen_functions.tree_obj_poll,
        update=gen_functions.tree_obj_update,
    )


class BiomeListItem(bpy.types.PropertyGroup):
    def _get_biome_settings_node(self):
        mod_group = get_mod_group(get_at_obj(bpy.context))
        for node in mod_group.nodes:
            if node.type == "GROUP" and node.node_tree == self.biome_settings:
                return node

    biome_settings_node = GetSetProperty(get=_get_biome_settings_node)

    biome_settings_enum: EnumProperty(
        items=get_biome_settings_enum,
        name="Biome settings",
        description="The biome settings to use for this biome slot",
        update=biome_settings_enum_update,
    )

    biome_settings: PointerProperty(
        type=bpy.types.GeometryNodeTree,
        name="Biome settings",
        description="The backend reference to the biome settings node group",
        update=biome_settings_update,
    )

    # display

    name: StringProperty(
        name="Name",
        description="Biome display name name",
        default="Biome",
    )

    show_viewport: BoolProperty(
        name="Show viewport",
        default=True,
        description="Show biome in the viewport",
        update=biome_show_viewport_update,
    )

    show_render: BoolProperty(
        name="Show rendered",
        default=True,
        description="Show biome when rendered",
        update=biome_show_render_update,
    )

    index: IntProperty(
        name="Index",
        description="The index of this item in the list",
        get=helpers.get_bl_list_index,
    )

    # misc

    distribution_type: EnumProperty(
        items=[
            ("DENSITY", "Density", "Controls how many trees are in an area"),
            ("SCALE", "Scale", "Controls how big the trees are in an area"),
        ]
    )

    density: PointerProperty(type=ATDensitySettings)
    scale: PointerProperty(type=ATScaleSettings)

    copy_to: BoolProperty(
        default=False,
    )


class TreeListItem(icons.PreviewSettingsParent):
    # prev_cache_small = []
    # prev_cache_large = []

    # display

    name: StringProperty(name="Name", description="Tree display name", default="Untitled")

    show_render: BoolProperty(
        name="Show rendered",
        default=True,
        description="Show tree when rendered",
        update=tree_show_render_update,
    )

    quality: EnumProperty(
        items=imp_functions.get_tree_quality_items,
        name="Quality",
        description="The resolution of this tree",
        update=quality_update,
    )

    selected_tree: EnumProperty(
        items=icons.get_tree_enum_previews,
        update=selected_tree_update,
    )

    # internal

    tree_obj: PointerProperty(
        type=bpy.types.Object,
    )

    icon_value: IntProperty(get=get_tree_icon_value)

    obj_name: StringProperty(
        name="tree_name",
        description="tree object name",
        default="Tree",
    )

    previous_tree: StringProperty(
        description="used to delete previous tree in selected tree update func",
        default="default.png",
    )

    copy_to: BoolProperty(
        default=False,
        description="Whether to copy settings to this tree",
    )


class CopyMenuItem(bpy.types.PropertyGroup):
    """This is a general purpose collection property, mainly used for copying
    properties and drawing properties for multiple trees"""

    name: StringProperty(
        name="Name",
        description="The name of this setting",
    )

    type: StringProperty(
        name="Type",
        description="The type of this setting",
    )

    subtype: StringProperty(
        name="Subtype",
        description="The subtype of this setting",
    )

    tags: StringProperty(
        name="Tags",
        description="The tags of this setting. Tags are separated by ', ' (comma, space)",
    )

    exclusive: BoolProperty(
        description="Whether to allow multiple items to be on at the same time",
        default=False,
    )

    hide: BoolProperty(
        default=False,
    )

    on: BoolProperty(
        default=True,
        update=copy_item_on_update,
    )

    icon: IntProperty(default=0)

    from_node_input: BoolProperty(
        default=False,
    )

    float: FloatProperty(default=0.5)

    string: StringProperty()

    color: FloatVectorProperty(size=4)

    bool: BoolProperty()

    object: PointerProperty(type=bpy.types.Object)

    collection: PointerProperty(type=bpy.types.Collection)


classes = (
    ATDensitySettings,
    ATScaleSettings,
    ObjectListItem,
    BiomeListItem,
    TreeListItem,
    ATGeneratorSettings,
    ATMaterialSettings,
    ATObjectSettings,
    ATNodeGroupSettings,
    CopyMenuItem,
)


def register():
    for cls in classes:
        bpy.utils.register_class(cls)

    bpy.types.Scene.alpha_trees = PointerProperty(type=ATGeneratorSettings)
    bpy.types.Object.at = PointerProperty(type=ATObjectSettings)
    bpy.types.GeometryNodeTree.at = PointerProperty(type=ATNodeGroupSettings)
    bpy.types.Material.at = PointerProperty(type=ATMaterialSettings)

    # lists
    ATGeneratorSettings.obj_list = CollectionProperty(type=ObjectListItem)
    ATGeneratorSettings.copy_items = CollectionProperty(type=CopyMenuItem)
    ATObjectSettings.biome_list = CollectionProperty(type=BiomeListItem)
    ATNodeGroupSettings.tree_list = CollectionProperty(type=TreeListItem)

    bpy.types.Image.at_quality = EnumProperty(
        items=(
            ("", "none", "none"),  # if created before update
            ("high", "high", "high"),
            ("med", "med", "med"),
            ("low", "low", "low"),
        )
    )

    bpy.types.Collection.at_scatter_obj_list_active_idx = IntProperty()

    # for some reason prefs get reset when the addon is reloaded using the VSCode plugin,
    # so these get set when reloading if on dev computer. Hacky, but it works.
    if IS_DEV:
        prefs = get_prefs()
        # prefs.lib_path = r"D:\Documents\Blender\Naturall\trees\Alpha Trees\Libraries\alpha_trees_lib_2_0_PRO"
        prefs.lib_path = r"D:\Documents\Blender\Naturall\trees\Alpha Trees\Libraries\Testing\small scale testing"
        prefs.show_dev_options = True


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)

    del bpy.types.Scene.alpha_trees
    del bpy.types.GeometryNodeTree.at
    del bpy.types.Material.at
    del bpy.types.Image.at_quality
